# from __future__ import print_function

import re
import os
import gzip
import time

import numpy as np


def _mylbl(d, x):
    return d.get(x, r"%s" % x)


class TwissPlot(object):
    lglabel = {
        "betx": r"$\beta_x$",
        "bety": r"$\beta_y$",
        "dx": r"$D_x$",
        "dy": r"$D_y$",
        "mux": r"$\mu_x$",
        "muy": r"$\mu_y$",
        "ax_chrom": "$A_x$",
        "ay_chrom": "$A_y$",
        "bx_chrom": "$B_x$",
        "by_chrom": "$B_y$",
        "wx_chrom": "$W_x$",
        "wy_chrom": "$W_y$",
        "sigx": r"$\sigma_x=\sqrt{\beta_x \epsilon}$",
        "sigy": r"$\sigma_y=\sqrt{\beta_y \epsilon}$",
        "sigdx": r"$\sigma_{D_x}=D_x \delta$",
        "n1": r"Aperture [$\sigma$]",
        "rad_int_hx": r"$\mathcal{H}_{x}$",
        "rad_int_hy": r"$\mathcal{H}_{y}$",
    }

    axlabel = {
        "s": r"$s [m]$",
        "ss": r"$s [m]$",
        "betx": r"$\beta [m]$",
        "bety": r"$\beta [m]$",
        "mux": r"$\mu/(2 \pi)$",
        "muy": r"$\mu/(2 \pi)$",
        "dx": r"$D [m]$",
        "dy": r"$D [m]$",
        "x": r"$co [m]$",
        "y": r"$co [m]$",
        "sigx": r"$\sigma$ [mm]",
        "sigy": r"$\sigma$ [mm]",
        "sigdx": r"$\sigma$ [mm]",
        "ax_chrom": "$A$",
        "ay_chrom": "$A$",
        "bx_chrom": "$B$",
        "by_chrom": "$B$",
        "wx_chrom": "$W$",
        "wy_chrom": "$W$",
        "n1": r"Aperture [$\sigma$]",
        "rad_int_hx": r"$\mathcal{H}$",
        "rad_int_hy": r"$\mathcal{H}$",
    }
    autoupdate = []

    def ani_autoupdate(self):
        from matplotlib.animation import FuncAnimation

        self._ani = FuncAnimation(self.figure, self.update, blit=False, interval=1000)

    def ani_stopupdate(self):
        del self._ani

    @classmethod
    def on_updated(cls, fun):
        cls.on_update = fun

    def __init__(
        self,
        table,
        x="",
        yl="",
        yr="",
        idx=slice(None),
        clist="k r b g c m",
        lattice=None,
        figure=None,
        figlabel=None,
        ax=None,
        axleft=None,
        axright=None,
        axlattice=None,
        hover=False,
        grid = True,
        figsize=(6.4*1.2, 4.8)
    ):

        import matplotlib.pyplot as plt

        yl, yr, clist = list(map(str.split, (yl, yr, clist)))
        #    timeit('Init',True)
        self.color = {}
        self.left = None
        self.right = None
        self.lattice = None
        self.pre = None
        self.grid = grid
        self.table, self.x, self.yl, self.yr, self.idx, self.clist = (
            table,
            x,
            yl,
            yr,
            idx,
            clist,
        )
        self.ax = ax
        self.used_ax = False

        if figure is not None:
            self.figure = figure

        if ax is not None:
            self.figure = ax.figure
        elif figure is None:
            self.figure = plt.figure(num=figlabel, figsize=figsize)

        if figlabel is not None:
            self.figure.clf()

        for i in self.yl + self.yr:
            self.color[i] = self.clist.pop(0)
            self.clist.append(self.color[i])

        if lattice and x == "s":
            self.lattice = self._new_axis(axlattice)
            # self.lattice.set_frame_on(False)
            #      self.lattice.set_autoscale_on(False)
            self.lattice.yaxis.set_visible(False)

        if yl:
            self.left = self._new_axis(axleft)
            #      self.left.set_autoscale_on(False)

        if yr:
            self.right = self._new_axis(axright)
            #      self.right.set_autoscale_on(False)
            self.left.yaxis.set_label_position("right")
            self.left.yaxis.set_ticks_position("right")

        #    timeit('Setup')
        self.run()
        if lattice and x == "s":
            self.lattice.set_autoscale_on(False)

        if yl:
            self.left.set_autoscale_on(False)
            self.left.yaxis.set_label_position("left")
            self.left.yaxis.set_ticks_position("left")

        if yr:
            self.right.set_autoscale_on(False)
            self.right.yaxis.set_label_position("right")
            self.right.yaxis.set_ticks_position("right")

        if hover:
            self.set_hover()

    #    timeit('Update')
    def _new_axis(self, ax=None):
        if self.ax is None:
            out = self.figure.add_subplot(111)
            self.figure.subplots_adjust(right=0.75)
            self.ax = out
        if self.used_ax:
            out = self.ax.twinx()
        else:
            out = self.ax
            self.used_ax = True
        return out

    def __repr__(self):
        return object.__repr__(self)

    def _trig(self):
        print("optics trig")
        self.run()

    def update(self, *args):
        if hasattr(self.table, "reload"):
            if self.table.reload():
                self.run()
                return self
        return False

    #  def _wx_callback(self,*args):
    #    self.update()
    #    wx.WakeUpIdle()
    #
    #  def autoupdate(self):
    #    if plt.rcParams['backend']=='WXAgg':
    #      wx.EVT_IDLE.Bind(wx.GetApp(),wx.ID_ANY,wx.ID_ANY,self._wx_callback)
    #    return self
    #
    #  def stop_update(self):
    #    if plt.rcParams['backend']=='WXAgg':
    #      wx.EVT_IDLE.Unbind(wx.GetApp(),wx.ID_ANY,wx.ID_ANY,self._callback)
    #
    #  def __del__(self):
    #    if hasattr(self,'_callback'):
    #      self.stop_update()

    def run(self):

        import matplotlib.pyplot as plt

        self.ont = self.table
        self.xaxis = self.ont[self.x][self.idx]
        self.lines = []
        self.legends = []
        self.names = []
        #    self.figure.lines=[]
        #    self.figure.patches=[]
        #    self.figure.texts=[]
        #    self.figure.images = []
        self.figure.legends = []

        if self.lattice:
            self.lattice.clear()
            self._lattice(["k0l", "kn0l", "angle"], "#a0ffa0", "Bend h")
            self._lattice(["ks0l"], "#ffa0a0", "Bend v")
            self._lattice(["kn1l", "k1l"], "#a0a0ff", "Quad")
            self._lattice(["hkick"], "#e0a0e0", "Kick h")
            self._lattice(["vkick"], "#a0e0e0", "Kick v")
            self._lattice(["kn2l", "k2l"], "#e0e0a0", "Sext")
        if self.left:
            self.left.clear()
            for i in self.yl:
                self._column(i, self.left, self.color[i])
        if self.right:
            self.right.clear()
            for i in self.yr:
                self._column(i, self.right, self.color[i])
        self.ax.set_xlabel(_mylbl(self.axlabel, self.x))
        self.ax.set_xlim(min(self.xaxis), max(self.xaxis))
        self.ax.legend(
            self.lines, self.legends, loc="upper right", bbox_to_anchor=(1.35, 1.)
        )
        self.ax.grid(self.grid)
        self.figure.canvas.draw()
        if hasattr(self, "on_run"):
            self.on_run(self)

    def set_hover(self):
        self.figure.canvas.mpl_connect("motion_notify_event", self.pick)

    def pick(self, event):
        for ii, ll in enumerate(self.lines):
            _, data = ll.contains(event)
            lgd = self.names[ii]
            if "ind" in data:
                xx = ll.get_xdata()
                yy = ll.get_ydata()
                for idx in data["ind"]:
                    if "name" in self.table._col_names:
                        name = self.table.name[idx]
                    else:
                        name = ""
                    if "element_type" in self.table._col_names:
                        elem_type = self.table.element_type[idx]
                    if not elem_type.startswith("Drift"):
                        print(f"{name:25}, s={xx[idx]:15.6g}, {lgd:>10}={yy[idx]:15.6g}")
        # pos = np.array([event.mouseevent.x, event.mouseevent.y])
        # name = event.artist.elemname
        # prop = event.artist.elemprop
        # value = event.artist.elemvalue
        # print("\n %s.%s=%s" % (name, prop, value), end=" ")

    #  def button_press(self,mouseevent):
    #    rel=np.array([mouseevent.x,mouseevent.y])
    #    dx,dy=self.pickpos/rel
    #    print 'release'
    #    self.t[self.pickname][self.pickprop]*=dy
    #    self.t.track()
    #    self.update()

    def _lattice(self, names, color, lbl):
        # print("start lattice %s" % names)
        #    timeit('start lattice %s' % names,1)
        import matplotlib.pyplot as plt

        vd = 0
        sp = self.lattice
        s = self.ont.s
        l = np.diff(s, append=[s[-1]])
        for name in names:
            myvd = self.ont._data.get(name, None)
            if myvd is not None:
                vdname = name
                vd = myvd[self.idx] + vd
        if np.any(vd != 0):
            m = np.abs(vd).max()
            if m > 1e-10:
                c = np.where(abs(vd) > m * 1e-4)[0]
                if len(c) > 0:
                    if np.all(l[c] > 0):
                        vd[c] = vd[c] / l[c]
                        m = abs(vd[c]).max()
                    vd[c] /= m
                    if self.ont._is_s_begin:
                        bplt = self.lattice.bar(
                            s[c] + l[c] / 2, vd[c], l[c], picker=True
                        )  # changed
                    else:
                        bplt = self.lattice.bar(
                            s[c] - l[c] / 2, vd[c], l[c], picker=True
                        )  # changed
                    plt.setp(bplt, facecolor=color, edgecolor=color)
                    if bplt:
                        self.lines.append(bplt[0])
                        self.legends.append(lbl)
                        self.names.append(lbl)
                    row_names = self.ont.name
                    for r, name in zip(bplt, c):
                        r.elemname = row_names[name]
                        r.elemprop = vdname
                        r.elemvalue = getattr(self.ont, vdname)[name]
                self.lattice.set_ylim(-1.5, 1.5)

    #    timeit('end lattice')

    def _column(self, name, sp, color):
        fig, s = self.figure, self.xaxis
        y = self.ont[name][self.idx]
        (bxp,) = sp.plot(s, y, color, label=_mylbl(self.lglabel, name))
        sp.set_ylabel(_mylbl(self.axlabel, name))
        self.lines.append(bxp)
        self.legends.append(_mylbl(self.lglabel, name))
        self.names.append(name)
        sp.autoscale_view()

    def savefig(self, name):
        self.figure.savefig(name)
        return self

    def ylim(
        self,
        left_lo=None,
        left_hi=None,
        right_lo=None,
        right_hi=None,
        lattice_lo=None,
        lattice_hi=None,
    ):
        if self.left is not None:
            lo, hi = self.left.get_ylim()
            if left_lo is None:
                left_lo = lo
            if left_hi is None:
                left_hi = hi
            self.left.set_ylim(left_lo, left_hi)
        if self.right is not None:
            lo, hi = self.right.get_ylim()
            if right_lo is None:
                right_lo = lo
            if right_hi is None:
                right_hi = hi
            self.right.set_ylim(right_lo, right_hi)
        if self.lattice is not None:
            lo, hi = self.lattice.get_ylim()
            if lattice_lo is None:
                lattice_lo = lo
            if lattice_hi is None:
                lattice_hi = hi
            self.lattice.set_ylim(lattice_lo, lattice_hi)
        return self

    def xlim(self, lo=None, hi=None):
        self.ax.set_xlim(lo, hi)
        return self

    def set_s_label(self, regexp="ip.*"):
        sel = self.table.rows[regexp]
        self.ax.set_xticks(sel.s, sel.name)
        self.ax.set_xlabel(None)
        return self

    def move_legend(self, left=0, bottom=0, width=0, height=0):
        """Uses ax.legend_.set_bbox_to_anchor"""
        self.ax.legend_.set_bbox_to_anchor((left, bottom, width, height))
        return self



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import xtrack as xt
from scipy.constants import c as clight

line = xt.Line.from_json('../../test_data/sps_w_spacecharge/line_no_spacecharge.json')
line.build_tracker()
tw0 = line.twiss4d()

f0 = 1 / (line.get_length()/(clight*line.particle_ref.beta0[0]))
h_rf = 4620

tt = line.get_table(attr=True)
tt_cav = tt.rows[tt.element_type=='Cavity']
tt_cav.cols['frequency voltage lag'].show()

line['dfreq'] = 1.e3
line['on_rf1'] = 1.
line['on_rf2'] = 1.

line['acta.31637'].absolute_time = 1
line['acta.31637'].frequency = h_rf * f0 + line.ref['dfreq']
line['acta.31637'].voltage = 1.2e6 * line.ref['on_rf1']
line['acta.31637'].lag = 180

line['actd.31934'].absolute_time = 1
line['actd.31934'].frequency = h_rf * f0 - line.ref['dfreq']
line['actd.31934'].voltage = 1.2e6 * line.ref['on_rf2']
line['actd.31934'].lag = 180

tt = line.get_table(attr=True)
tt_cav = tt.rows[tt.element_type=='Cavity']
tt_cav.cols['frequency voltage lag'].show()

line['on_rf2'] = 0.
tw_rf1 = line.twiss(search_for_t_rev=True)
line['on_rf2'] = 1.

line['on_rf1'] = 0.
tw_rf2 = line.twiss(search_for_t_rev=True)
line['on_rf1'] = 1.

particles = line.build_particles(
    method='4d',
    x_norm=0, y_norm=0,
    delta=np.linspace(-8e-3, 8e-3, 1000))

particles.t_sim = line.get_length() / line.particle_ref._xobject.beta0[0] / clight
line.track(particles, num_turns=1000, turn_by_turn_monitor=True, with_progress=10)
rec = line.record_last_track

# match a bunch (no frequency shift)
dfreq_tmp = line['dfreq']
line['dfreq'] = 0
line['on_rf2'] = 0.
import xpart as xp
# bunch0 = xp.generate_matched_gaussian_bunch(sigma_z=0.15, num_particles=500,
#                                             nemitt_x=0, nemitt_y=0, line=line)
bucket_length = line.get_length() / h_rf
train0 = xp.generate_matched_gaussian_multibunch_beam(
    sigma_z=0.1, nemitt_x=0, nemitt_y=0, line=line,
    bunch_intensity_particles=1e11,
    bunch_num_particles=500, filling_scheme=5*[1, 0, 0, 0, 0])
line['on_rf2'] = 1.
line['dfreq'] = dfreq_tmp

train1 = train0.copy()
train1.delta += tw_rf1.delta[0]
train2 = train0.copy()
train2.delta += tw_rf2.delta[0]
train2.zeta += 20 * bucket_length

two_trains = xt.Particles.merge([train1, train2])

num_turns = 400
line.track(two_trains, num_turns=num_turns, with_progress=10,
           turn_by_turn_monitor=True)
mon = line.record_last_track

import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1)
for ii in range(rec.x.shape[0]):
    mask = rec.state[ii, :]>0
    plt.plot(rec.zeta[ii, mask], rec.delta[ii, mask])

plt.grid(linestyle='--')
plt.xlabel('z [m]')
plt.ylabel(r'$\Delta p / p_0$')

plt.figure(2)
plt.plot(rec.zeta.T)

plt.figure(3)
plt.plot(two_trains.zeta, two_trains.delta, '.')

i_turn = 350
plt.figure(4)
plt.plot(mon.zeta.T[i_turn, :], mon.delta.T[i_turn, :], '.')

plt.show()

# def update_plot(i_turn):
#     plt.clf()
#     plt.plot(mon.zeta.T[i_turn, :], mon.delta.T[i_turn, :], '.', markersize=1)
#     plt.xlim(-60, 60)
#     plt.ylim(-12e-3, 12e-3)
#     plt.xlabel('z [m]')
#     plt.ylabel(r'$\Delta p / p_0$')
#     plt.title(f'Turn {i_turn}')
#     plt.subplots_adjust(left=0.2)
#     plt.grid(alpha=0.5)


# import matplotlib.animation as animation
# fig = plt.figure()
# animation_fig = animation.FuncAnimation(
#     fig, update_plot, frames=range(0, num_turns, 3))
# animation_fig.save('slipstack.gif', fps=10)

def update_plot(i_turn, fig):
    plt.clf()
    plt.plot(mon.zeta.T[i_turn, :], mon.delta.T[i_turn, :], '.', markersize=1)
    plt.xlim(-40, 40)
    plt.ylim(-12e-3, 12e-3)
    plt.xlabel('z [m]')
    plt.ylabel(r'$\Delta p / p_0$')
    plt.title(f'Turn {i_turn}')
    plt.subplots_adjust(left=0.2)
    plt.grid(alpha=0.5)

fig = plt.figure()
from matplotlib.animation import FFMpegFileWriter
moviewriter = FFMpegFileWriter(fps=15)
with moviewriter.saving(fig, 'slipstack.mp4', dpi=100):
    for j in range(0, num_turns, 2):
        update_plot(j, fig)
        moviewriter.grab_frame()



# def update_plot(i_turn):
#     plt.clf()
#     plt.plot(mon.zeta.T[i_turn, :], mon.y.T[i_turn, :], '.', markersize=1)
#     plt.xlim(-60, 60)
#     plt.ylim(-5e-3, 5e-3)
#     plt.xlabel('z [m]')
#     plt.ylabel(r'y [mm]')
#     plt.title(f'Turn {i_turn}')
#     plt.subplots_adjust(left=0.2)
#     plt.grid(alpha=0.5)

# import matplotlib.animation as animation
# fig = plt.figure()
# animation_fig = animation.FuncAnimation(
#     fig, update_plot, frames=range(0, num_turns, 1))
# animation_fig.save('slipstack_y.gif', fps=30)




import numpy as np
from cpymad.madx import Madx

import xobjects as xo
import xtrack as xt
import xpart as xp

# Choose a context
ctx = xo.ContextCpu()

# Import SPS lattice and build a tracker
mad = Madx()
seq_name = 'sps'
mad.call('../../test_data/sps_w_spacecharge/sps_thin.seq')
mad.use(seq_name)
madtw = mad.twiss()

line = xt.Line.from_madx_sequence(mad.sequence[seq_name])
line.particle_ref = xt.Particles(p0c=400e9, mass0=xt.PROTON_MASS_EV)
line.build_tracker(_context=ctx)

# Switch on RF and twiss
line['acta.31637'].voltage = 7e6
line['acta.31637'].lag = 180.
twxt = line.twiss()

# Get revolution period
T_rev = twxt['T_rev0']

# Extract list of elements to trim (all focusing quads)
tt = line.get_table()
elements_to_trim = tt.rows[tt.element_type == 'Multipole'].rows['qf.*'].name
# => contains ['qf.52010', 'qf.52210', 'qf.52410', 'qf.52610', 'qf.52810',
#              'qf.53010', 'qf.53210', 'qf.53410', 'qf.60010', 'qf.60210', ...]

# Build a custom setter
qf_setter = xt.MultiSetter(line, elements_to_trim,
                            field='knl', index=1 # we want to change knl[1]
                            )

# Get the initial values of the quad strength
k1l_0 = qf_setter.get_values()

# Generate particles to be tracked
# (we choose to match the distribution without accounting for spacecharge)
particles = xp.generate_matched_gaussian_bunch(_context=ctx,
         num_particles=100, total_intensity_particles=1e10,
         nemitt_x=3e-6, nemitt_y=3e-6, sigma_z=15e-2,
         line=line)

# Define amplitude and phase of the quadrupole ripple
f_quad = 50. # Hz
A_quad = 0.01 # relative amplitude

# Track the particles and apply turn-by-turn change to all selected quads
num_turns = 2000

check_trim  = []

for ii in range(num_turns):
    if ii % 100 == 0: print(f'Turn {ii} of {num_turns}')

    # Change the strength of the quads
    k1l = k1l_0 * (1 + A_quad * np.sin(2*np.pi*f_quad*ii*T_rev))
    qf_setter.set_values(k1l)

    # Track one turn
    line.track(particles)

    # Log the strength of one quad to check
    check_trim.append(ctx.nparray_from_context_array(line['qf.52010'].knl)[1])

# Plot the evolution of the quad strength
import matplotlib.pyplot as plt
plt.plot(check_trim)
plt.show()







import numpy as np
import xtrack as xt

# Define elements
pi = np.pi
lbend = 3
# Build a simple ring
env = xt.Environment()
line = env.new_line(components=[
    env.new('mqf.1', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.1',  xt.Drift, length=1),
    env.new('mb1.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.1',  xt.Drift, length=1),

    env.new('mqd.1', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.1',  xt.Drift, length=1),
    env.new('mb2.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.1',  xt.Drift, length=1),

    env.new('mqf.2', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.2',  xt.Drift, length=1),
    env.new('mb1.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.2',  xt.Drift, length=1),

    env.new('mqd.2', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.2',  xt.Drift, length=1),
    env.new('mb2.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d4.2',  xt.Drift, length=1),
])

#!start-doc-part
hundred_cuts = np.linspace(0, line.get_length(), num=100)
print(f'Make cuts every {hundred_cuts[1]} metres')  # => 21 centimetres
line.cut_at_s(s=list(hundred_cuts))

line.get_table()
# returns:
#
# Table: 139 rows, 7 cols
# name                    s element_type         isthick isreplica parent_name ...
# mqf.1_entry             0 Marker                 False     False        None
# mqf.1..0                0 ThickSliceQuadrupole    True     False       mqf.1
# mqf.1..1         0.214141 ThickSliceQuadrupole    True     False       mqf.1
# mqf.1_exit            0.3 Marker                 False     False        None
# d1.1..0               0.3 DriftSlice              True     False        d1.1
# d1.1..1          0.428283 DriftSlice              True     False        d1.1
# d1.1..2          0.642424 DriftSlice              True     False        d1.1
# d1.1..3          0.856566 DriftSlice              True     False        d1.1
# d1.1..4           1.07071 DriftSlice              True     False        d1.1
# d1.1..5           1.28485 DriftSlice              True     False        d1.1
# mb1.1_entry           1.3 Marker                 False     False        None
# mb1.1..entry_map      1.3 ThinSliceBendEntry     False     False       mb1.1
# mb1.1..0              1.3 ThickSliceBend          True     False       mb1.1
# mb1.1..1          1.49899 ThickSliceBend          True     False       mb1.1
# mb1.1..2          1.71313 ThickSliceBend          True     False       mb1.1
# mb1.1..3          1.92727 ThickSliceBend          True     False       mb1.1
# mb1.1..4          2.14141 ThickSliceBend          True     False       mb1.1
# etc...



import xtrack as xt

lhc = xt.Environment.from_madx('job.madx')


# mad = Madx()
# mad.call(fname)

# kwargs = dict()

# lines = {}
# for nn in mad.sequence.keys():
#     lines[nn] = xt.Line.from_madx_sequence(
#         mad.sequence[nn],
#         allow_thick=True,
#         deferred_expressions=True,
#         **kwargs)

#     lines[nn].particle_ref = xt.Particles(
#         mass0=mad.sequence[nn].beam.mass*1e9,
#         q0=mad.sequence[nn].beam.charge,
#         gamma0=mad.sequence[nn].beam.gamma)

#     if mad.sequence[nn].beam.bv == -1:
#         lines[nn].twiss_default['reverse'] = True

# multiline = xt.Environment(lines=lines)


import numpy as np
import xtrack as xt

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')

tw = line.twiss(method='4d')

mng = line.to_madng(sequence_name='lhcb1')

mng["mytwtable", 'mytwflow'] = mng.twiss(
    sequence=mng.lhcb1, method=4, mapdef=2, implicit=True, nslice=3, save="'atbody'")

print(mng["mytwtable"].mu1[-1])
assert np.isclose(mng["mytwtable"].mu1[-1][0], 62.31, atol=1e-6, rtol=0)


mng.send('''
    local track in MAD  -- like "from MAD import track"
    local mytrktable, mytrkflow = MAD.track{sequence=MADX.lhcb1, method=4,
                                            mapdef=4, nslice=3}

        -- print(MAD.typeid.is_damap(mytrkflow[1]))

    local normal in MAD.gphys  -- like "from MAD.gphys import normal"
    local my_norm_for = normal(mytrkflow[1]):analyse('anh') -- anh stands for anharmonicity

    local nf = my_norm_for
    py:send({
            nf:q1{1}, -- qx from the normal form (fractional part)
            nf:q2{1}, -- qy
            nf:dq1{1}, -- dqx / d delta
            nf:dq2{1}, -- dqy / d delta
            nf:dq1{2}, -- d2 qx / d delta2
            nf:dq2{2}, -- d2 qy / d delta2
            nf:anhx{1, 0}, -- dqx / d(2 jx)
            nf:anhy{0, 1}, -- dqy / d(2 jy)
            nf:anhx{0, 1}, -- dqx / d(2 jy)
            nf:anhy{1, 0}, -- dqy / d(2 jx)
            })
''')

out = mng.recv()

print(f'''
    q1: {out[0]}
    q2: {out[1]}
    dq1: {out[2]}
    dq2: {out[3]}
    d2q1: {out[4]}
    d2q2: {out[5]}
    anhx: {out[6]}
    anhy: {out[7]}
    anhx: {out[8]}
    anhy: {out[9]}
''')





from cpymad.madx import Madx

mad = Madx()

mad.input('''

set, format="19.15f";
option,update_from_parent=true;

radiationandrf = 1;

call,file="fccee_t.seq";

pbeam =   182.5;
exbeam = 1.46e-9;
eybeam = 2.9e-12;
nbun =    48;
npar =   2.3e11;
halfcrossingangle = 0.015;
ebeam = sqrt( pbeam^2 + emass^2 );
beam, particle=positron, npart=npar, kbunch=nbun, energy=ebeam, radiate=false, bv=+1, ex=exbeam, ey=eybeam;

!rf and radiation off
use, sequence = fccee_p_ring;
voltca1save = voltca1;
voltca2save = voltca2;


!rf and radiation off

voltca1 = 0;
voltca2 = 0;
SAVEBETA, LABEL=B.IP, PLACE=#s, SEQUENCE=FCCEE_P_RING;
''')

tw_0 = mad.twiss(chrom=True).dframe()
summ_0 = mad.table.summ.dframe()


mad.input('''
! radiation on

voltca1 = voltca1save;
voltca2 = voltca2save;
beam, radiate=true;

use, sequence = fccee_p_ring;


! tapering and matching
MATCH, sequence=FCCEE_P_RING, BETA0 = B.IP, tapering;
  VARY, NAME=LAGCA1, step=1.0E-7;
  VARY, NAME=LAGCA2, step=1.0E-7;
  CONSTRAINT, SEQUENCE=FCCEE_P_RING, RANGE=#e, PT=0.0;
  JACOBIAN,TOLERANCE=1.0E-14, CALLS=3000;
ENDMATCH;

twiss, BETA0 = B.IP, tolerance=1e-12;

MATCH, sequence=FCCEE_P_RING, BETA0 = B.IP, tapering;
  VARY, NAME=LAGCA1, step=1.0E-7;
  VARY, NAME=LAGCA2, step=1.0E-7;
  CONSTRAINT, SEQUENCE=FCCEE_P_RING, RANGE=#e, PT=0.0;
  JACOBIAN,TOLERANCE=1.0E-14, CALLS=3000;
ENDMATCH;

use, sequence = fccee_p_ring;
''')

tw_1 = mad.twiss(chrom=True).dframe()
summ_1 = mad.table.summ.dframe()

import matplotlib.pyplot as plt
plt.close('all')

# Plot beta beating
plt.figure(1)
ax1 = plt.subplot(211)
plt.plot(tw_0['s'], tw_1['betx']/tw_0['betx']-1)
plt.ylabel(r'$\Delta\beta_x/\beta_x$')
plt.subplot(212, sharex=ax1)
plt.plot(tw_0['s'], tw_1['bety']/tw_0['bety']-1)
plt.ylabel(r'$\Delta\beta_y/\beta_y$')
plt.xlabel('s [m]')

plt.show()


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from matplotlib import pyplot as plt

import xobjects as xo
import xtrack as xt

context = xo.ContextCpu()

# ######################################### #
# Create a line with detuning               #
# and the corresponding tracker             #
# ######################################### #
Q_x = .28
Q_y = .31
Q_s = 1E-3
det_xx = 1E6
det_xy = -7E5
det_yy = -5E5
det_yx = 1E5
element = xt.LineSegmentMap(_context=context,
        qx=Q_x, det_xx = det_xx, det_xy = det_xy,
        qy=Q_y, det_yy = det_yy, det_yx = det_yx,
        qs = Q_s,bets=1.0)
line = xt.Line(elements = [element])
line.particle_ref = xt.Particles(p0c=7000e9, mass0=xt.PROTON_MASS_EV)
line.build_tracker()
# ######################################### #
# Create footprint on a uniform grid        #
# in transverse actions Jx,Jy               #
# (given the large number of points, it     #
#  is advised not to keep the FFTs in       #
#  memory)                                  #
# ######################################### #
nemitt_x = 2.5e-6
nemitt_y = 4.5e-6
footprint = line.get_footprint(
        nemitt_x=nemitt_x, nemitt_y=nemitt_y,
        mode='uniform_action_grid',n_x_norm=200,n_y_norm=210,keep_fft=False)
# ######################################### #
# Draw the stability diagram by computing   #
# the dispersion integral for a Gaussian    #
# beam in X and Y for a set of coherent     #
# tunes with vanishing imaginary part       #
# ######################################### #
tune_shifts_x,tune_shifts_y = footprint.get_stability_diagram(context)
# ######################################### #
# Plot                                      #
# ######################################### #
plt.figure(0)
plt.plot(np.real(tune_shifts_x),np.imag(tune_shifts_x),'-b',label='Horizontal')
plt.plot(np.real(tune_shifts_y),np.imag(tune_shifts_y),'-g',label='Vertical')
plt.xlabel(r'$\Re\Delta Q$')
plt.ylabel(r'$\Im\Delta Q$')
plt.legend()
plt.show()




import numpy as np
from cpymad.madx import Madx
import xtrack as xt

# line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')


line = xt.Environment.from_json(
    '../../test_data/hllhc15_collider/collider_00_from_mad.json').lhcb1
# Rotations not supported in thin
for nn in list(line.element_names):
    if (nn.startswith('mb') and
       (nn.endswith('_tilt_entry') or nn.endswith('_tilt_exit'))):
        line.element_names.remove(nn)

line.build_tracker()
mad_seq = line.to_madx_sequence(sequence_name='myseq')

mad = Madx()
mad.input(mad_seq)
mad.beam(particle='proton', energy=7000e9) #!!!!!
mad.use('myseq')

line2 = xt.Line.from_madx_sequence(mad.sequence.myseq, deferred_expressions=True)
line2.particle_ref = line.particle_ref

for ll in [line, line2]:
    ll.vars['vrf400'] = 16 # Check voltage expressions
    ll.vars['lagrf400.b1'] = 0.52 # Check lag expressions
    ll.vars['on_x1'] = 100 # Check kicker expressions
    ll.vars['on_sep2'] = 2 # Check kicker expressions
    ll.vars['on_x5'] = 123 # Check kicker expressions
    ll.vv['kqtf.b1'] += 1e-5 # Check quad expressions
    ll.vv['ksf.b1'] += 1e-3  # Check sext expressions
    ll.vv['kqs.l4b1'] += 1e-4 # Check skew expressions
    ll.vv['kof.a34b1'] = 3 # Check oct expressions
    ll.vars['on_crab1'] = -190 # Check cavity expressions
    ll.vars['on_crab5'] = -130 # Check cavity expressions

tw = line.twiss()
tw2 = line2.twiss()

assert np.all(tw2.rows['ip.*'].name == tw.rows['ip.*'].name)

assert np.allclose(tw2.rows['ip.*'].s, tw.rows['ip.*'].s, rtol=0, atol=2e-9)
assert np.allclose(tw2.rows['ip.*'].x, tw.rows['ip.*'].x, rtol=0, atol=1e-9)
assert np.allclose(tw2.rows['ip.*'].y, tw.rows['ip.*'].y, rtol=0, atol=1e-9)
assert np.allclose(tw2.rows['ip.*'].px, tw.rows['ip.*'].px, rtol=0, atol=1e-9)
assert np.allclose(tw2.rows['ip.*'].py, tw.rows['ip.*'].py, rtol=0, atol=1e-9)
assert np.allclose(tw2.rows['ip.*'].mux, tw.rows['ip.*'].mux, rtol=0, atol=1e-8)
assert np.allclose(tw2.rows['ip.*'].muy, tw.rows['ip.*'].muy, rtol=0, atol=1e-8)
assert np.allclose(tw2.rows['ip.*'].betx, tw.rows['ip.*'].betx, rtol=1e-7, atol=0)
assert np.allclose(tw2.rows['ip.*'].bety, tw.rows['ip.*'].bety, rtol=1e-7, atol=0)
assert np.allclose(tw2.rows['ip.*'].ax_chrom, tw.rows['ip.*'].ax_chrom, rtol=1e-5, atol=0)
assert np.allclose(tw2.rows['ip.*'].dx_zeta, tw.rows['ip.*'].dx_zeta, rtol=2e-4, atol=0)
assert np.allclose(tw2.rows['ip.*'].dy_zeta, tw.rows['ip.*'].dy_zeta, rtol=2e-4, atol=0)

assert np.isclose(tw2.qs, tw.qs, rtol=0, atol=1e-7)
assert np.isclose(tw2.ddqx, tw.ddqx, rtol=2e-4, atol=0)


import numpy as np
import matplotlib.pyplot as plt

v = 1.
l_ref = 1

t_ref = l_ref/v

l = np.array([0.95, 1, 1.05])

t_max = 80
t = np.linspace(0, t_max, 10000)

s = np.zeros((len(l), len(t)))
for il, ll in enumerate(l):
    s[il, :] = np.mod(0.62*l_ref + v * t * ll/l_ref, l_ref)

passing = np.diff(s, prepend=0) < 0

import matplotlib.pyplot as plt
plt.close('all')

plt.figure(1, figsize=(6.4*1.8*0.8, 4.8*0.8))
sp1 = plt.subplot(311)
sp2 = plt.subplot(312, sharex=sp1)
sp3 = plt.subplot(313, sharex=sp1)

sp1.stem(t, passing[1, :], markerfmt=' ', basefmt='C0', linefmt='C0')
sp2.stem(t, passing[0, :], markerfmt=' ', basefmt='C1', linefmt='C1')
sp3.stem(t, passing[2, :], markerfmt=' ', basefmt='C2', linefmt='C2')

t_turn = np.arange(0, t_max, t_ref)
for tt in t_turn:
    for sp in [sp1, sp2, sp3]:
        sp.axvline(tt, color='k', linestyle='--', alpha=0.4)

sp3.set_xlim(0, t_max/t_ref)
sp3.set_xlabel(r'$t~/~T_0$')

for sp in [sp1, sp2, sp3]:
    sp.set_ylim(0, 1.1)

plt.subplots_adjust(bottom=.14, top=.95, hspace=0.3)

# For zoom
# plt.subplots_adjust(right=.6)

plt.show()




import pathlib

from cpymad.madx import Madx

import xdeps as xd
import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts
from xtrack.slicing import Teapot, Strategy

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_madloader_lhc_thick(test_context):
    mad = Madx(stdout=False)

    mad.input(f"""
    call,file="{str(test_data_folder)}/hllhc15_thick/lhc.seq";
    call,file="{str(test_data_folder)}/hllhc15_thick/hllhc_sequence.madx";
    seqedit,sequence=lhcb1;flatten;cycle,start=IP7;flatten;endedit;
    seqedit,sequence=lhcb2;flatten;cycle,start=IP7;flatten;endedit;
    beam, sequence=lhcb1, particle=proton, pc=7000;
    call,file="{str(test_data_folder)}/hllhc15_thick/opt_round_150_1500.madx";
    """)

    mad.use(sequence="lhcb1")
    seq = mad.sequence.lhcb1
    mad.twiss()

    line = xt.Line.from_madx_sequence(mad.sequence.lhcb1,
                allow_thick=True, deferred_expressions=True)
    line.particle_ref = xp.Particles(mass0=seq.beam.mass*1e9, gamma0=seq.beam.gamma)
    line.twiss_default['method'] = '4d'
    line.twiss_default['matrix_stability_tol'] = 100


    line.build_tracker(_context=test_context)

    # To have the very same model
    for ee in seq.elements:
        if hasattr(ee, 'KILL_ENT_FRINGE'):
            ee.KILL_ENT_FRINGE = True
        if hasattr(ee, 'KILL_EXI_FRINGE'):
            ee.KILL_EXI_FRINGE = True

    line.configure_bend_model(edge='suppressed')

    tw0 = line.twiss()
    twmad = mad.twiss(table='twiss')
    tmad = xd.Table(twmad)

    print(f'dqx xsuite:      {tw0.dqx}')
    print(f'dqx mad nochrom: {twmad.summary.dq1}')
    print(f'dqy xsuite:      {tw0.dqy}')
    print(f'dqy mad nochrom: {twmad.summary.dq2}')

    xo.assert_allclose(tw0.dqx, twmad.summary.dq1, atol=0.2, rtol=0)
    xo.assert_allclose(tw0.dqy, twmad.summary.dq2, atol=0.2, rtol=0)
    xo.assert_allclose(tw0.qx, twmad.summary.q1, atol=1e-6, rtol=0)
    xo.assert_allclose(tw0.qy, twmad.summary.q2, atol=1e-6, rtol=0)
    xo.assert_allclose(tw0['betx', 'ip1'], tmad['betx', 'ip1:1'], atol=0, rtol=1e-5)
    xo.assert_allclose(tw0['bety', 'ip1'], tmad['bety', 'ip1:1'], atol=0, rtol=1e-5)
    xo.assert_allclose(tw0['betx', 'ip5'], tmad['betx', 'ip5:1'], atol=0, rtol=1e-5)
    xo.assert_allclose(tw0['bety', 'ip5'], tmad['bety', 'ip5:1'], atol=0, rtol=1e-5)


@for_all_test_contexts
def test_slicing_lhc_thick(test_context):
    line = xt.Line.from_json(test_data_folder /
            'hllhc15_thick/lhc_thick_with_knobs.json')
    line.twiss_default['method'] = '4d'
    line.build_tracker(_context=test_context)

    line_thick = line.copy()
    line_thick.build_tracker(_context=test_context)

    slicing_strategies = [
        Strategy(slicing=Teapot(1)),  # Default catch-all as in MAD-X
        Strategy(slicing=Teapot(4), element_type=xt.Bend),
        Strategy(slicing=Teapot(20), element_type=xt.Quadrupole),
        Strategy(slicing=Teapot(2), name=r'^mb\..*'),
        Strategy(slicing=Teapot(5), name=r'^mq\..*'),
        Strategy(slicing=Teapot(2), name=r'^mqt.*'),
        Strategy(slicing=Teapot(60), name=r'^mqx.*'),
    ]

    line.discard_tracker()
    line.slice_thick_elements(slicing_strategies=slicing_strategies)
    line.build_tracker(_context=test_context)

    tw = line.twiss()
    tw_thick = line_thick.twiss()

    beta_beat_x_at_ips = [tw['betx', f'ip{nn}'] / tw_thick['betx', f'ip{nn}'] - 1
                            for nn in range(1, 9)]
    beta_beat_y_at_ips = [tw['bety', f'ip{nn}'] / tw_thick['bety', f'ip{nn}'] - 1
                            for nn in range(1, 9)]

    xo.assert_allclose(beta_beat_x_at_ips, 0, atol=3e-3)
    xo.assert_allclose(beta_beat_y_at_ips, 0, atol=3e-3)

    # Checks on orbit knobs
    xo.assert_allclose(tw_thick['px', 'ip1'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw_thick['py', 'ip1'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw_thick['px', 'ip5'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw_thick['py', 'ip5'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw['px', 'ip1'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw['py', 'ip1'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw['px', 'ip5'], 0, rtol=0, atol=1e-7)
    xo.assert_allclose(tw['py', 'ip5'], 0, rtol=0, atol=1e-7)

    line.vars['on_x1'] = 50
    line.vars['on_x5'] = 60
    line_thick.vars['on_x1'] = 50
    line_thick.vars['on_x5'] = 60

    tw = line.twiss()
    tw_thick = line_thick.twiss()

    xo.assert_allclose(tw_thick['px', 'ip1'], 50e-6, rtol=0, atol=5e-7)
    xo.assert_allclose(tw_thick['py', 'ip1'], 0, rtol=0, atol=5e-7)
    xo.assert_allclose(tw_thick['px', 'ip5'], 0, rtol=0, atol=5e-7)
    xo.assert_allclose(tw_thick['py', 'ip5'], 60e-6, rtol=0, atol=5e-7)
    xo.assert_allclose(tw['px', 'ip1'], 50e-6, rtol=0, atol=5e-7)
    xo.assert_allclose(tw['py', 'ip1'], 0, rtol=0, atol=5e-7)
    xo.assert_allclose(tw['px', 'ip5'], 0, rtol=0, atol=5e-7)
    xo.assert_allclose(tw['py', 'ip5'], 60e-6, rtol=0, atol=5e-7)

    line_thick.match(
        vary=[
            xt.Vary('kqtf.b1', step=1e-8),
            xt.Vary('kqtd.b1', step=1e-8),
            xt.Vary('ksf.b1', step=1e-8),
            xt.Vary('ksd.b1', step=1e-8),
        ],
        targets = [
            xt.Target('qx', 62.27, tol=1e-4),
            xt.Target('qy', 60.29, tol=1e-4),
            xt.Target('dqx', 10.0, tol=0.05),
            xt.Target('dqy', 12.0, tol=0.05)])

    line.match(
        vary=[
            xt.Vary('kqtf.b1', step=1e-8),
            xt.Vary('kqtd.b1', step=1e-8),
            xt.Vary('ksf.b1', step=1e-8),
            xt.Vary('ksd.b1', step=1e-8),
        ],
        targets = [
            xt.Target('qx', 62.27, tol=1e-4),
            xt.Target('qy', 60.29, tol=1e-4),
            xt.Target('dqx', 10.0, tol=0.05),
            xt.Target('dqy', 12.0, tol=0.05)])

    tw = line.twiss()
    tw_thick = line_thick.twiss()

    xo.assert_allclose(tw_thick.qx, 62.27, rtol=0, atol=1e-4)
    xo.assert_allclose(tw_thick.qy, 60.29, rtol=0, atol=1e-4)
    xo.assert_allclose(tw_thick.dqx, 10.0, rtol=0, atol=0.05)
    xo.assert_allclose(tw_thick.dqy, 12.0, rtol=0, atol=0.05)
    xo.assert_allclose(tw.qx, 62.27, rtol=0, atol=1e-4)
    xo.assert_allclose(tw.qy, 60.29, rtol=0, atol=1e-4)
    xo.assert_allclose(tw.dqx, 10.0, rtol=0, atol=0.05)
    xo.assert_allclose(tw.dqy, 12.0, rtol=0, atol=0.05)

    xo.assert_allclose(line.vars['kqtf.b1']._value, line_thick.vars['kqtf.b1']._value,
                        rtol=0.03, atol=0)
    xo.assert_allclose(line.vars['kqtd.b1']._value, line_thick.vars['kqtd.b1']._value,
                        rtol=0.03, atol=0)
    xo.assert_allclose(line.vars['ksf.b1']._value, line_thick.vars['ksf.b1']._value,
                        rtol=0.03, atol=0)
    xo.assert_allclose(line.vars['ksd.b1']._value, line_thick.vars['ksd.b1']._value,
                        rtol=0.03, atol=0)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt
import xobjects as xo


# In this case the element internal record is made of two separate tables each
# with its own record index.

class Table1(xo.HybridClass):
    _xofields = {
        '_index': xt.RecordIndex,
        'particle_x': xo.Float64[:],
        'particle_px': xo.Float64[:],
        'at_element': xo.Int64[:],
        'at_turn': xo.Int64[:],
        'particle_id': xo.Int64[:]
        }

class Table2(xo.HybridClass):
    _xofields = {
        '_index': xt.RecordIndex,
        'generated_rr': xo.Float64[:],
        'at_element': xo.Int64[:],
        'at_turn': xo.Int64[:],
        'particle_id': xo.Int64[:]
        }

class TestElementRecord(xo.HybridClass):
    _xofields = {
        'table1': Table1,
        'table2': Table2
        }

# The two tables in the internal record can be accessed independently in the C
# code of the beam element.

TestElement_track_method_source = r'''
    /*gpufun*/
    void TestElement_track_local_particle(TestElementData el, LocalParticle* part0){

        // Extract the record and record_index
        TestElementRecordData record = TestElementData_getp_internal_record(el, part0);
        Table1Data table1 = NULL;
        Table2Data table2 = NULL;
        RecordIndex table1_index = NULL;
        RecordIndex table2_index = NULL;
        if (record){
            table1 = TestElementRecordData_getp_table1(record);
            table2 = TestElementRecordData_getp_table2(record);
            table1_index = Table1Data_getp__index(table1);
            table2_index = Table2Data_getp__index(table2);
        }

        int64_t n_kicks = TestElementData_get_n_kicks(el);
        printf("n_kicks %d\n", (int)n_kicks);

        //start_per_particle_block (part0->part)

            // Record in table1 info about the ingoing particle
            if (record){
                // Get a slot in table1
                int64_t i_slot = RecordIndex_get_slot(table1_index);
                // The returned slot id is negative if record is NULL or if record is full
                printf("i_slot %d\n", (int)i_slot);

                if (i_slot>=0){
                        Table1Data_set_particle_x(table1, i_slot,
                                                    LocalParticle_get_x(part));
                        Table1Data_set_particle_px(table1, i_slot,
                                                    LocalParticle_get_px(part));
                        Table1Data_set_at_element(table1, i_slot,
                                                    LocalParticle_get_at_element(part));
                        Table1Data_set_at_turn(table1, i_slot,
                                                    LocalParticle_get_at_turn(part));
                        Table1Data_set_particle_id(table1, i_slot,
                                                    LocalParticle_get_particle_id(part));
                }
            }

            for (int64_t i = 0; i < n_kicks; i++) {
                double rr = 1e-6 * RandomUniform_generate(part);
                LocalParticle_add_to_px(part, rr);

                // Record in table2 info about the generated kicks
                if (record){
                    // Get a slot in table2
                    int64_t i_slot = RecordIndex_get_slot(table2_index);
                    // The returned slot id is negative if record is NULL or if record is full
                    printf("i_slot %d\n", (int)i_slot);

                    if (i_slot>=0){
                            Table2Data_set_generated_rr(table2, i_slot, rr);
                            Table2Data_set_at_element(table2, i_slot,
                                                        LocalParticle_get_at_element(part));
                            Table2Data_set_at_turn(table2, i_slot,
                                                        LocalParticle_get_at_turn(part));
                            Table2Data_set_particle_id(table2, i_slot,
                                                        LocalParticle_get_particle_id(part));
                    }
                }
            }

        //end_per_particle_block
    }
    '''

# To allow elements of a given type to store data in a structure of the type defined
# above we need to add in the element class an attribute called
# `_internal_record_class` to which we bind the data structure type defined above.

class TestElement(xt.BeamElement):
    _xofields={
        'n_kicks': xo.Int64,
        }

    _internal_record_class = TestElementRecord

    _depends_on = [xt.RandomUniform]

    _extra_c_sources = [TestElement_track_method_source]

# Once these steps are done, the TestElement and its recording feature are ready
# and can be used as follows.

# Line, tracker, particles can be created as usual
line=xt.Line(elements = [
    xt.Drift(length=1.), TestElement(n_kicks=10),
    xt.Drift(length=1.), TestElement(n_kicks=5)])
line.build_tracker()
line._needs_rng = True # Test elements use the random number generator
part = xt.Particles(p0c=6.5e12, x=[1e-3,2e-3,3e-3])

# The record is by default disabled and can be enabled using the following
# dedicated method of the line object.
# The capacity allocated for the two tables needs to be provided in a dictionary:
line.start_internal_logging_for_elements_of_type(
                                                    TestElement,
                                    capacity={'table1': 5000, 'table2': 10000})

# Track!
line.track(part, num_turns=10)

# We can now inspect the two tables in the `record`, e.g `record.table1.particle_x`,
# `record.table2.generated_rr`. The number of used slots in each can be found in
# the corresponding index object e.g. record.table1._index.num_recorded.

# The recording can be stopped with the following method (optionally reinitializing
# the internal buffers to free memot=ry):
line.stop_internal_logging_for_all_elements(reinitialize_io_buffer=True)

# Track more turns (without logging information in `record`)
line.track(part, num_turns=10)


#!end-doc-part




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xtrack as xt
import xobjects as xo

context = xo.ContextCpu()

part_gen_range = 0.11
n_part=100000

particles = xt.Particles(_context=context,
        p0c=6500e9,
        x=np.random.uniform(-part_gen_range, part_gen_range, n_part),
        y=np.random.uniform(-part_gen_range, part_gen_range, n_part))

aper = xt.LimitRacetrack(_context=context,
                         min_x=-5e-2, max_x=10e-2,
                         min_y=-2e-2, max_y=4e-2,
                         a=2e-2, b=1e-2)

aper.track(particles)

part_state = context.nparray_from_context_array(particles.state)
part_x = context.nparray_from_context_array(particles.x)
part_y = context.nparray_from_context_array(particles.y)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(part_x, part_y, '.', color='red')
plt.plot(part_x[part_state>0], part_y[part_state>0], '.', color='green')


xy_out = np.array([
    [-4.8e-2, 3.7e-2],
    [9.6e-2, 3.7e-2],
    [-4.5e-2, -1.8e-2],
    [9.8e-2, -1.8e-2],
    ])

xy_in = np.array([
    [-4.2e-2, 3.3e-2],
    [9.4e-2, 3.6e-2],
    [-3.8e-2, -1.8e-2],
    [9.2e-2, -1.8e-2],
    ])

xy_all = np.concatenate([xy_out, xy_in], axis=0)

particles = xt.Particles(_context=context,
        p0c=6500e9,
        x=xy_all[:, 0],
        y=xy_all[:, 1])

aper.track(particles)

part_state = context.nparray_from_context_array(particles.state)
part_x = context.nparray_from_context_array(particles.x)
part_y = context.nparray_from_context_array(particles.y)
part_id = context.nparray_from_context_array(particles.particle_id)

assert np.all(part_state[part_id<4] == 0)
assert np.all(part_state[part_id>=4] == 1)

plt.plot(part_x, part_y, '.', color='k')
plt.plot(part_x[part_state>0], part_y[part_state>0], 'x', color='k')



plt.show()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

context = xo.ContextCpu()

q_x_set = .28
q_y_set = .31
q_s_set = .01
# Test branches with and without detuning
for chrm in [1e-5, 0]:
    el = xt.LineSegmentMap(_context=context,
            qx=q_x_set, qy=q_y_set, qs=q_s_set,
            bets=800.,
            dqx=chrm
            )

    part = xt.Particles(_context=context, x=[1], y=[1], zeta=[1],
                        p0c=6500e9)

    n_turns = 1024
    x_record = []
    y_record = []
    z_record = []
    for _ in range(n_turns):
        x_record.append(part.x[0])
        y_record.append(part.y[0])
        z_record.append(part.zeta[0])
        el.track(part)

    import nafflib

    q_x_meas = nafflib.get_tune(np.array(x_record))
    q_y_meas = nafflib.get_tune(np.array(y_record))
    q_s_meas = nafflib.get_tune(np.array(z_record))

    assert np.isclose(q_x_meas, q_x_set, rtol=1e-10, atol=1e-6)
    assert np.isclose(q_y_meas, q_y_set, rtol=1e-10, atol=1e-6)
    assert np.isclose(q_s_meas, q_s_set, rtol=1e-10, atol=1e-6)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import pathlib

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_match_and_track_from_element(test_context):
    # Load machine model (from pymask)
    filename = test_data_folder.joinpath('lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.build_tracker(_context=test_context)
    line.reset_s_at_end_turn = False

    assert not line.iscollective
    line.particle_ref = xp.Particles.from_dict(input_data['particle'])

    # Check matching of a one-sigma circle in ip2
    r_sigma = 1
    theta = np.linspace(0, 2*np.pi, 1000)

    at_element = 'ip2'
    particles = line.build_particles(
                    x_norm=r_sigma*np.cos(theta), px_norm=r_sigma*np.sin(theta),
                    nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                    at_element=at_element)

    tw = line.twiss(at_elements=[at_element])

    particles.move(_context=xo.context_default) # To easily do the checks with numpy
    xo.assert_allclose(
        np.sqrt(tw['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=1e-3, atol=0)
    particles.move(_context=test_context)

    # Check that tracking starts from the right place
    line.track(particles, turn_by_turn_monitor='ONE_TURN_EBE')
    mon = line.record_last_track
    i_ele_start = line.element_names.index(at_element)
    assert np.all(mon.at_element[:, :i_ele_start] == 0)
    assert np.all(mon.at_element[:, i_ele_start] == i_ele_start)
    assert np.all(mon.at_element[:, -1] == len(line.element_names))

    # Check that distribution is matched at the end of the turn
    tw0 = line.twiss(at_elements=[0])
    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw0['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=2e-3, atol=0)

    # Check multiple turns
    at_element = 'ip2'
    particles = line.build_particles(
                    x_norm=r_sigma*np.cos(theta), px_norm=r_sigma*np.sin(theta),
                    nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                    at_element=at_element)

    tw = line.twiss(at_elements=[at_element])

    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=1e-3, atol=0)
    particles.move(_context=test_context)

    line.track(particles, num_turns=3)

    tw0 = line.twiss(at_elements=[0])
    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw0['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=2e-3, atol=0)
    assert np.all(particles.at_turn==3)
    xo.assert_allclose(particles.s, 3*line.get_length(), rtol=0, atol=1e-7)

    # Check collective case
    line_w_collective = xt.Line.from_dict(input_data['line'])
    for ip in range(8):
        line_w_collective.element_dict[f'ip{ip+1}'].iscollective = True
        line_w_collective.element_dict[f'ip{ip+1}'].move(_context=test_context)
    line_w_collective.build_tracker(_context=test_context)
    line_w_collective.reset_s_at_end_turn = False

    assert line_w_collective.iscollective
    line_w_collective.particle_ref = xp.Particles.from_dict(input_data['particle'])
    assert len(line_w_collective.tracker._parts) == 16

    at_element = 'ip2'
    particles = line_w_collective.build_particles(
                    x_norm=r_sigma*np.cos(theta), px_norm=r_sigma*np.sin(theta),
                    nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                    at_element=at_element)

    tw = line_w_collective.twiss(at_elements=[at_element])

    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=1e-3, atol=0)
    particles.move(_context=test_context)

    line_w_collective.track(particles, num_turns=3)

    tw0 = line_w_collective.twiss(at_elements=[0])
    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw0['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=3e-3, atol=0)
    assert np.all(particles.at_turn==3)
    xo.assert_allclose(particles.s, 3*line_w_collective.get_length(), rtol=0, atol=1e-7)

    # Check match_at_s
    at_element = 'ip6'
    particles = line_w_collective.build_particles(
                    x_norm=r_sigma*np.cos(theta), px_norm=r_sigma*np.sin(theta),
                    nemitt_x=2.5e-6, nemitt_y=2.5e-6,
                    at_element=at_element,
                    match_at_s=line_w_collective.get_s_position('ip6') + 100
                    )

    tw = line_w_collective.twiss(at_elements=[at_element])

    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=1e-3, atol=0)
    xo.assert_allclose(particles.s, tw['s'][0], atol=1e-8, rtol=0)

    phasex_first_part = np.angle(particles.x[0] / np.sqrt(tw['betx'][0]) -
                1j*(particles.x[0]  * tw['alfx'][0] / np.sqrt(tw['betx'][0]) +
                        particles.px[0] * np.sqrt(tw['betx'][0])))

    mu_at_s = line_w_collective.twiss(
        at_s=line_w_collective.get_s_position('ip6') + 100)['mux'][0]
    mu_at_element = line_w_collective.twiss(at_elements=[at_element])['mux'][0]

    xo.assert_allclose(phasex_first_part, (mu_at_element - mu_at_s)*2*np.pi,
                    atol=0, rtol=0.02)
    particles.move(_context=test_context)

    line_w_collective.track(particles, num_turns=3)

    tw0 = line_w_collective.twiss(at_elements=[0])
    particles.move(_context=xo.context_default)
    xo.assert_allclose(
        np.sqrt(tw0['betx'][0]*2.5e-6/particles.beta0[0]/particles.gamma0[0]),
        np.max(np.abs(particles.x - np.mean(particles.x))), rtol=2e-3, atol=0)
    assert np.all(particles.at_turn==3)
    xo.assert_allclose(particles.s, 3*line_w_collective.get_length(), rtol=0, atol=1e-7)



import pathlib

import numpy as np
import pytest
from cpymad.madx import Madx

import xdeps as xd
import xobjects as xo
import xtrack as xt
import xtrack._temp.lhc_match as lm
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

@pytest.mark.parametrize(
    'config',
    ['noshift', 'shift']
)
@for_all_test_contexts
@fix_random_seed(2836475)
def test_lhc_match_phase_15(test_context, config):

    if config == 'noshift':
        d_mux_15_b1 = 0
        d_muy_15_b1 = 0
        d_mux_15_b2 = 0
        d_muy_15_b2 = 0
    elif config == 'shift':
        d_mux_15_b1 = 0.07
        d_muy_15_b1 = 0.05
        d_mux_15_b2 = -0.1
        d_muy_15_b2 = -0.12
    else:
        raise ValueError(f'Invalid config {config}')

    staged_match = True

    collider = xt.Environment.from_json(
        test_data_folder / 'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(_context=test_context)
    collider.vars.load_madx_optics_file(
        test_data_folder / "hllhc15_thick/opt_round_150_1500.madx")

    # to have no rematching w.r.t. madx
    default_tol = {None: 1e-8, 'betx': 5e-6, 'bety': 5e-6, 'dx': 1e-7,
                  'alfx': 5e-6, 'alfy': 5e-6}

    lm.set_var_limits_and_steps(collider)

    tw0 = collider.twiss()

    optimizers = {}

    print('Matching ip15 phase:')
    opt = lm.change_phase_non_ats_arcs(collider,
        d_mux_15_b1=d_mux_15_b1, d_muy_15_b1=d_muy_15_b1,
        d_mux_15_b2=d_mux_15_b2, d_muy_15_b2=d_muy_15_b2,
        solve=True, default_tol=default_tol)
    optimizers['phase_15'] = opt

    arc_periodic_solution = lm.get_arc_periodic_solution(collider)

    optimizers.update({'b1':{}, 'b2':{}})

    for bn in ['b1', 'b2']:

        line_name = f'lhc{bn}'

        muxip1_l = collider.varval[f'muxip1{bn}_l']
        muyip1_l = collider.varval[f'muyip1{bn}_l']
        muxip1_r = collider.varval[f'muxip1{bn}_r']
        muyip1_r = collider.varval[f'muyip1{bn}_r']

        muxip5_l = collider.varval[f'muxip5{bn}_l']
        muyip5_l = collider.varval[f'muyip5{bn}_l']
        muxip5_r = collider.varval[f'muxip5{bn}_r']
        muyip5_r = collider.varval[f'muyip5{bn}_r']

        muxip2 = collider.varval[f'muxip2{bn}']
        muyip2 = collider.varval[f'muyip2{bn}']
        muxip4 = collider.varval[f'muxip4{bn}']
        muyip4 = collider.varval[f'muyip4{bn}']
        muxip6 = collider.varval[f'muxip6{bn}']
        muyip6 = collider.varval[f'muyip6{bn}']
        muxip8 = collider.varval[f'muxip8{bn}']
        muyip8 = collider.varval[f'muyip8{bn}']

        mux12 = collider.varval[f'mux12{bn}']
        muy12 = collider.varval[f'muy12{bn}']
        mux45 = collider.varval[f'mux45{bn}']
        muy45 = collider.varval[f'muy45{bn}']
        mux56 = collider.varval[f'mux56{bn}']
        muy56 = collider.varval[f'muy56{bn}']
        mux81 = collider.varval[f'mux81{bn}']
        muy81 = collider.varval[f'muy81{bn}']

        betx_ip1 = collider.varval[f'betxip1{bn}']
        bety_ip1 = collider.varval[f'betyip1{bn}']
        betx_ip5 = collider.varval[f'betxip5{bn}']
        bety_ip5 = collider.varval[f'betyip5{bn}']

        betx_ip2 = collider.varval[f'betxip2{bn}']
        bety_ip2 = collider.varval[f'betyip2{bn}']

        alfx_ip3 = collider.varval[f'alfxip3{bn}']
        alfy_ip3 = collider.varval[f'alfyip3{bn}']
        betx_ip3 = collider.varval[f'betxip3{bn}']
        bety_ip3 = collider.varval[f'betyip3{bn}']
        dx_ip3 = collider.varval[f'dxip3{bn}']
        dpx_ip3 = collider.varval[f'dpxip3{bn}']
        mux_ir3 = collider.varval[f'muxip3{bn}']
        muy_ir3 = collider.varval[f'muyip3{bn}']

        alfx_ip4 = collider.varval[f'alfxip4{bn}']
        alfy_ip4 = collider.varval[f'alfyip4{bn}']
        betx_ip4 = collider.varval[f'betxip4{bn}']
        bety_ip4 = collider.varval[f'betyip4{bn}']
        dx_ip4 = collider.varval[f'dxip4{bn}']
        dpx_ip4 = collider.varval[f'dpxip4{bn}']

        alfx_ip6 = collider.varval[f'alfxip6{bn}']
        alfy_ip6 = collider.varval[f'alfyip6{bn}']
        betx_ip6 = collider.varval[f'betxip6{bn}']
        bety_ip6 = collider.varval[f'betyip6{bn}']
        dx_ip6 = collider.varval[f'dxip6{bn}']
        dpx_ip6 = collider.varval[f'dpxip6{bn}']

        alfx_ip7 = collider.varval[f'alfxip7{bn}']
        alfy_ip7 = collider.varval[f'alfyip7{bn}']
        betx_ip7 = collider.varval[f'betxip7{bn}']
        bety_ip7 = collider.varval[f'betyip7{bn}']
        dx_ip7 = collider.varval[f'dxip7{bn}']
        dpx_ip7 = collider.varval[f'dpxip7{bn}']
        mux_ir7 = collider.varval[f'muxip7{bn}']
        muy_ir7 = collider.varval[f'muyip7{bn}']

        alfx_ip8 = collider.varval[f'alfxip8{bn}']
        alfy_ip8 = collider.varval[f'alfyip8{bn}']
        betx_ip8 = collider.varval[f'betxip8{bn}']
        bety_ip8 = collider.varval[f'betyip8{bn}']
        dx_ip8 = collider.varval[f'dxip8{bn}']
        dpx_ip8 = collider.varval[f'dpxip8{bn}']

        tw_sq_a81_ip1_a12 = lm.propagate_optics_from_beta_star(collider, ip_name='ip1',
                line_name=f'lhc{bn}', start=f's.ds.r8.{bn}', end=f'e.ds.l2.{bn}',
                beta_star_x=betx_ip1, beta_star_y=bety_ip1)

        tw_sq_a45_ip5_a56 = lm.propagate_optics_from_beta_star(collider, ip_name='ip5',
                line_name=f'lhc{bn}', start=f's.ds.r4.{bn}', end=f'e.ds.l6.{bn}',
                beta_star_x=betx_ip5, beta_star_y=bety_ip5)

        (mux_ir2_target, muy_ir2_target, mux_ir4_target, muy_ir4_target,
        mux_ir6_target, muy_ir6_target, mux_ir8_target, muy_ir8_target
            ) = lm.compute_ats_phase_advances_for_auxiliary_irs(line_name,
                tw_sq_a81_ip1_a12, tw_sq_a45_ip5_a56,
                muxip1_l, muyip1_l, muxip1_r, muyip1_r,
                muxip5_l, muyip5_l, muxip5_r, muyip5_r,
                muxip2, muyip2, muxip4, muyip4, muxip6, muyip6, muxip8, muyip8,
                mux12, muy12, mux45, muy45, mux56, muy56, mux81, muy81)

        print(f"Matching IR2 {bn}")
        opt = lm.rematch_ir2(collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=tw_sq_a81_ip1_a12,
                    boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['23'],
                    mux_ir2=mux_ir2_target, muy_ir2=muy_ir2_target,
                    betx_ip2=betx_ip2, bety_ip2=bety_ip2,
                    solve=True, staged_match=staged_match,
                    default_tol=default_tol)
        optimizers[bn]['ir2'] = opt

        print(f"Matching IR3 {bn}")
        opt = lm.rematch_ir3(collider=collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['23'],
                    boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['34'],
                    mux_ir3=mux_ir3, muy_ir3=muy_ir3,
                    alfx_ip3=alfx_ip3, alfy_ip3=alfy_ip3,
                    betx_ip3=betx_ip3, bety_ip3=bety_ip3,
                    dx_ip3=dx_ip3, dpx_ip3=dpx_ip3,
                    solve=True, staged_match=staged_match, default_tol=default_tol)
        optimizers[bn]['ir3'] = opt

        print(f"Matching IR4 {bn}")
        opt = lm.rematch_ir4(collider=collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['34'],
                    boundary_conditions_right=tw_sq_a45_ip5_a56,
                    mux_ir4=mux_ir4_target, muy_ir4=muy_ir4_target,
                    alfx_ip4=alfx_ip4, alfy_ip4=alfy_ip4,
                    betx_ip4=betx_ip4, bety_ip4=bety_ip4,
                    dx_ip4=dx_ip4, dpx_ip4=dpx_ip4,
                    solve=True, staged_match=staged_match, default_tol=default_tol)
        optimizers[bn]['ir4'] = opt

        print(f"Matching IP6 {bn}")
        opt = lm.rematch_ir6(collider=collider, line_name=f'lhc{bn}',
                    boundary_conditions_left=tw_sq_a45_ip5_a56,
                    boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['67'],
                    mux_ir6=mux_ir6_target, muy_ir6=muy_ir6_target,
                    alfx_ip6=alfx_ip6, alfy_ip6=alfy_ip6,
                    betx_ip6=betx_ip6, bety_ip6=bety_ip6,
                    dx_ip6=dx_ip6, dpx_ip6=dpx_ip6,
                    solve=True, staged_match=staged_match, default_tol=default_tol)
        optimizers[bn]['ir6'] = opt

        print(f"Matching IP7 {bn}")
        opt = lm.rematch_ir7(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['67'],
                boundary_conditions_right=arc_periodic_solution[f'lhc{bn}']['78'],
                mux_ir7=mux_ir7, muy_ir7=muy_ir7,
                alfx_ip7=alfx_ip7, alfy_ip7=alfy_ip7,
                betx_ip7=betx_ip7, bety_ip7=bety_ip7,
                dx_ip7=dx_ip7, dpx_ip7=dpx_ip7,
                solve=True, staged_match=staged_match, default_tol=default_tol)
        optimizers[bn]['ir7'] = opt

        print(f"Matching IP8 {bn}")
        opt = lm.rematch_ir8(collider=collider, line_name=f'lhc{bn}',
                boundary_conditions_left=arc_periodic_solution[f'lhc{bn}']['78'],
                boundary_conditions_right=tw_sq_a81_ip1_a12,
                mux_ir8=mux_ir8_target, muy_ir8=muy_ir8_target,
                alfx_ip8=alfx_ip8, alfy_ip8=alfy_ip8,
                betx_ip8=betx_ip8, bety_ip8=bety_ip8,
                dx_ip8=dx_ip8, dpx_ip8=dpx_ip8,
                solve=True, staged_match=staged_match, default_tol=default_tol)
        optimizers[bn]['ir8'] = opt

    if config == 'noshift':
        # Check that the match of the optics did not touch anything
        opts_to_check = []
        opts_to_check.append(optimizers['phase_15']['opt_mq'])
        opts_to_check.append(optimizers['phase_15']['opt_mqt_b1'])
        opts_to_check.append(optimizers['phase_15']['opt_mqt_b2'])
        opts_to_check += optimizers['b1'].values()
        opts_to_check += optimizers['b2'].values()

        for opt in opts_to_check:
            ll = opt.log()
            assert len(ll) >= 2 # will be more than two for staged matches
            assert np.all(ll.vary[-1] == ll.vary[0])

    opt = lm.match_orbit_knobs_ip2_ip8(collider)
    optimizers['orbit_knobs'] = opt

    # Generate madx optics file
    lm.gen_madx_optics_file_auto(collider, f'opt_round_150_1500_xs_{config}.madx')

    tw = collider.twiss()

    # Tunes
    print('Tunes:')
    print(f"  b1: qx={tw.lhcb1.qx:6f} qy={tw.lhcb1.qy:6f}")
    print(f"  b2: qx={tw.lhcb2.qx:6f} qy={tw.lhcb2.qy:6f}")

    print('IP15 phase before:')
    print(f"  b1: d_mux={tw0.lhcb1['mux', 'ip5'] - tw0.lhcb1['mux', 'ip1']:6f} "
        f"      d_muy={tw0.lhcb1['muy', 'ip5'] - tw0.lhcb1['muy', 'ip1']:6f} ")
    print(f"  b2: d_mux={tw0.lhcb2['mux', 'ip5'] - tw0.lhcb2['mux', 'ip1']:6f} "
        f"      d_muy={tw0.lhcb2['muy', 'ip5'] - tw0.lhcb2['muy', 'ip1']:6f} ")

    print('IP15 phase after:')
    print(f"  b1: d_mux={tw.lhcb1['mux', 'ip5'] - tw.lhcb1['mux', 'ip1']:6f} "
        f"      d_muy={tw.lhcb1['muy', 'ip5'] - tw.lhcb1['muy', 'ip1']:6f} ")
    print(f"  b2: d_mux={tw.lhcb2['mux', 'ip5'] - tw.lhcb2['mux', 'ip1']:6f} "
        f"      d_muy={tw.lhcb2['muy', 'ip5'] - tw.lhcb2['muy', 'ip1']:6f} ")

    print('IP15 phase shifts:')
    print(f"  b1: d_mux={tw.lhcb1['mux', 'ip5'] - tw0.lhcb1['mux', 'ip5']:6f} "
                f"d_muy={tw.lhcb1['muy', 'ip5'] - tw0.lhcb1['muy', 'ip5']:6f} ")
    print(f"  b2: d_mux={tw.lhcb2['mux', 'ip5'] - tw0.lhcb2['mux', 'ip5']:6f} "
                f"d_muy={tw.lhcb2['muy', 'ip5'] - tw0.lhcb2['muy', 'ip5']:6f} ")

    collider_ref = xt.Environment.from_json(
        test_data_folder / 'hllhc15_thick/hllhc15_collider_thick.json')
    collider_ref.build_trackers()
    collider_ref.vars.load_madx_optics_file(
        test_data_folder / "hllhc15_thick/opt_round_150_1500.madx")

    tw = collider.twiss()
    tw_ref = collider_ref.twiss()

    d_mux = {'lhcb1': d_mux_15_b1, 'lhcb2': d_mux_15_b2}
    d_muy = {'lhcb1': d_muy_15_b1, 'lhcb2': d_muy_15_b2}

    for ll in ['lhcb1', 'lhcb2']:

        xo.assert_allclose(tw[ll].qx, tw_ref[ll].qx, atol=1e-7, rtol=0)
        xo.assert_allclose(tw[ll].qy, tw_ref[ll].qy, atol=1e-7, rtol=0)

        mux_15 = tw[ll]['mux', 'ip5'] - tw[ll]['mux', 'ip1']
        mux_ref_15 = tw_ref[ll]['mux', 'ip5'] - tw_ref[ll]['mux', 'ip1']

        muy_15 = tw[ll]['muy', 'ip5'] - tw[ll]['muy', 'ip1']
        muy_ref_15 = tw_ref[ll]['muy', 'ip5'] - tw_ref[ll]['muy', 'ip1']

        xo.assert_allclose(mux_15, mux_ref_15 + d_mux[ll], atol=1e-7, rtol=0)
        xo.assert_allclose(muy_15, muy_ref_15 + d_muy[ll], atol=1e-7, rtol=0)

        twip = tw[ll].rows['ip.*']
        twip_ref = tw_ref[ll].rows['ip.*']

        xo.assert_allclose(twip['betx'], twip_ref['betx'], rtol=1e-6, atol=0)
        xo.assert_allclose(twip['bety'], twip_ref['bety'], rtol=1e-6, atol=0)
        xo.assert_allclose(twip['alfx'], twip_ref['alfx'], rtol=0, atol=1e-5)
        xo.assert_allclose(twip['alfy'], twip_ref['alfy'], rtol=0, atol=1e-5)
        xo.assert_allclose(twip['dx'], twip_ref['dx'], rtol=0, atol=1e-6)
        xo.assert_allclose(twip['dy'], twip_ref['dy'], rtol=0, atol=1e-6)
        xo.assert_allclose(twip['dpx'], twip_ref['dpx'], rtol=0, atol=1e-7)
        xo.assert_allclose(twip['dpy'], twip_ref['dpy'], rtol=0, atol=1e-7)


    # -----------------------------------------------------------------------------

    # Check higher level knobs

    collider.vars['on_x2'] = 34
    tw = collider.twiss()
    collider.vars['on_x2'] = 0

    xo.assert_allclose(tw.lhcb1['py', 'ip2'], 34e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip2'], -34e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-8, rtol=0)

    collider.vars['on_x8'] = 35
    tw = collider.twiss()
    collider.vars['on_x8'] = 0

    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 35e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -35e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-8, rtol=0)

    collider.vars['on_sep2'] = 0.5
    tw = collider.twiss()
    collider.vars['on_sep2'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip2'], -0.5e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip2'], 0.5e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-8, rtol=0)

    collider.vars['on_sep8'] = 0.6
    tw = collider.twiss()
    collider.vars['on_sep8'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip8'], 0.6e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip8'], -0.6e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-8, rtol=0)

    # Check lower level knobs (disconnects higher level knobs)

    collider.vars['on_o2v'] = 0.3
    tw = collider.twiss()
    collider.vars['on_o2v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip2'], 0.3e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip2'], 0.3e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip2'], 0., atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip2'], 0., atol=1e-8, rtol=0)

    collider.vars['on_o2h'] = 0.4
    tw = collider.twiss()
    collider.vars['on_o2h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip2'], 0.4e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip2'], 0.4e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 0., atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], 0., atol=1e-8, rtol=0)

    collider.vars['on_o8v'] = 0.5
    tw = collider.twiss()
    collider.vars['on_o8v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip8'], 0.5e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip8'], 0.5e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 0., atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], 0., atol=1e-8, rtol=0)

    collider.vars['on_o8h'] = 0.6
    tw = collider.twiss()
    collider.vars['on_o8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0.6e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0.6e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 0., atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], 0., atol=1e-8, rtol=0)

    collider.vars['on_a2h'] = 20
    tw = collider.twiss()
    collider.vars['on_a2h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 20e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], 20e-6, atol=1e-8, rtol=0)

    collider.vars['on_a2v'] = 15
    tw = collider.twiss()
    collider.vars['on_a2v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip2'], 15e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip2'], 15e-6, atol=1e-8, rtol=0)

    collider.vars['on_a8h'] = 20
    tw = collider.twiss()
    collider.vars['on_a8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 20e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], 20e-6, atol=1e-8, rtol=0)

    collider.vars['on_a8v'] = 50
    tw = collider.twiss()
    collider.vars['on_a8v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 50e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], 50e-6, atol=1e-8, rtol=0)

    collider.vars['on_x2v'] = 100
    tw = collider.twiss()
    collider.vars['on_x2v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip2'], 100e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip2'], -100e-6, atol=1e-8, rtol=0)

    collider.vars['on_x2h'] = 120
    tw = collider.twiss()
    collider.vars['on_x2h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 120e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], -120e-6, atol=1e-8, rtol=0)


    collider.vars['on_x8h'] = 100
    tw = collider.twiss()
    collider.vars['on_x8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 100e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -100e-6, atol=1e-8, rtol=0)

    collider.vars['on_x8v'] = 120
    tw = collider.twiss()
    collider.vars['on_x8v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 120e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], -120e-6, atol=1e-8, rtol=0)

    collider.vars['on_sep2h'] = 1.6
    tw = collider.twiss()
    collider.vars['on_sep2h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip2'], 1.6e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip2'], -1.6e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip2'], 0, atol=1e-8, rtol=0)

    collider.vars['on_sep2v'] = 1.7
    tw = collider.twiss()
    collider.vars['on_sep2v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip2'], 1.7e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip2'], -1.7e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip2'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip2'], 0, atol=1e-8, rtol=0)

    collider.vars['on_sep8h'] = 1.5
    tw = collider.twiss()
    collider.vars['on_sep8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 1.5e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], -1.5e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], 0, atol=1e-8, rtol=0)

    collider.vars['on_sep8v'] = 1.7
    tw = collider.twiss()
    collider.vars['on_sep8v'] = 0

    xo.assert_allclose(tw.lhcb1['y', 'ip8'], 1.7e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['y', 'ip8'], -1.7e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['py', 'ip8'], 0, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['py', 'ip8'], 0, atol=1e-8, rtol=0)

    # Both knobs together
    collider.vars['on_x8h'] = 120
    collider.vars['on_sep8h'] = 1.7
    tw = collider.twiss()
    collider.vars['on_x8h'] = 0
    collider.vars['on_sep8h'] = 0

    xo.assert_allclose(tw.lhcb1['x', 'ip8'], 1.7e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['x', 'ip8'], -1.7e-3, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb1['px', 'ip8'], 120e-6, atol=1e-8, rtol=0)
    xo.assert_allclose(tw.lhcb2['px', 'ip8'], -120e-6, atol=1e-8, rtol=0)

    # Check generated optics in madx

    mad=Madx(stdout=False)
    mad.call(str(test_data_folder / 'hllhc15_thick/lhc.seq'))
    mad.call(str(test_data_folder / 'hllhc15_thick/hllhc_sequence.madx'))
    mad.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
    mad.use('lhcb1')
    mad.input('beam, sequence=lhcb2, particle=proton, energy=7000, bv=-1;')
    mad.use('lhcb2')
    mad.call(f"opt_round_150_1500_xs_{config}.madx")

    mad.input('twiss, sequence=lhcb1, table=twb1')
    mad.input('twiss, sequence=lhcb2, table=twb2')
    twmad_b1 = xd.Table(mad.table.twb1)
    twmad_b2 = xd.Table(mad.table.twb2)

    xo.assert_allclose(twmad_b1['betx', 'ip1:1'], 0.15, rtol=1e-7, atol=0)
    xo.assert_allclose(twmad_b1['bety', 'ip1:1'], 0.15, rtol=1e-7, atol=0)
    xo.assert_allclose(twmad_b2['betx', 'ip1:1'], 0.15, rtol=1e-7, atol=0)
    xo.assert_allclose(twmad_b2['bety', 'ip1:1'], 0.15, rtol=1e-7, atol=0)

    twmad_b1.rows['ip.*'].cols['betx bety x y px py'].show()
    twmad_b2.rows['ip.*'].cols['betx bety x y px py'].show()

    # Test orbit knobs
    mad.globals.on_x8 = 100
    mad.globals.on_x2 = 110

    mad.input('twiss, sequence=lhcb1, table=twb1')
    mad.input('twiss, sequence=lhcb2, table=twb2')
    twmad_b1 = xd.Table(mad.table.twb1)
    twmad_b2 = xd.Table(mad.table.twb2)

    xo.assert_allclose(twmad_b1['px', 'ip8:1'], 100e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twmad_b2['px', 'ip8:1'], -100e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(twmad_b1['py', 'ip2:1'], 110e-6, rtol=0, atol=5e-10)
    xo.assert_allclose(twmad_b2['py', 'ip2:1'], -110e-6, rtol=0, atol=5e-10)

    # Match tunes and chromaticity in the Xsuite model
    opt = collider.match(
        solve=False,
        vary=[
            xt.VaryList(['kqtf.b1', 'kqtd.b1', 'ksf.b1', 'ksd.b1'], step=1e-7),
            xt.VaryList(['kqtf.b2', 'kqtd.b2', 'ksf.b2', 'ksd.b2'], step=1e-7)],
        targets = [
            xt.TargetSet(line='lhcb1', qx=62.315, qy=60.325, tol=1e-6),
            xt.TargetSet(line='lhcb1', dqx=10.0, dqy=12.0, tol=1e-4),
            xt.TargetSet(line='lhcb2', qx=62.316, qy=60.324, tol=1e-6),
            xt.TargetSet(line='lhcb2', dqx=9.0, dqy=11.0, tol=1e-4)])
    opt.solve()

    # Transfer knobs to madx model and check matched values

    for kk, vv in opt.get_knob_values().items():
        mad.globals[kk] = vv

    mad.input('twiss, sequence=lhcb1, table=twb1')
    mad.input('twiss, sequence=lhcb2, table=twb2')

    xo.assert_allclose(mad.table.twb1.summary.q1, 62.315, rtol=0, atol=1e-6)
    xo.assert_allclose(mad.table.twb1.summary.q2, 60.325, rtol=0, atol=1e-6)
    xo.assert_allclose(mad.table.twb2.summary.q1, 62.316, rtol=0, atol=1e-6)
    xo.assert_allclose(mad.table.twb2.summary.q2, 60.324, rtol=0, atol=1e-6)
    xo.assert_allclose(mad.table.twb1.summary.dq1, 10.0, rtol=0, atol=0.5)
    xo.assert_allclose(mad.table.twb1.summary.dq2, 12.0, rtol=0, atol=0.5)
    xo.assert_allclose(mad.table.twb2.summary.dq1, 9.0, rtol=0, atol=0.5)
    xo.assert_allclose(mad.table.twb2.summary.dq2, 11.0, rtol=0, atol=0.5)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import copy

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_random_generation(test_context):

    part = xp.Particles(_context=test_context, p0c=6.5e12, x=[1,2,3])
    part._init_random_number_generator()

    class TestElement(xt.BeamElement):
        _xofields={
            'dummy': xo.Float64,
            }

        _depends_on = [xt.RandomNormal]

        _extra_c_sources = [
            '''
                /*gpufun*/
                void TestElement_track_local_particle(
                        TestElementData el, LocalParticle* part0){
                    //start_per_particle_block (part0->part)
                        double rr = RandomNormal_generate(part);
                        LocalParticle_set_x(part, rr);
                    //end_per_particle_block
                }
            '''
        ]

    telem = TestElement(_context=test_context)

    telem.track(part)

    # Use turn-by-turn monitor to acquire some statistics
    line = xt.Line(elements=[telem])
    line.build_tracker(_buffer=telem._buffer)

    line.track(part, num_turns=1e6, turn_by_turn_monitor=True)

    for i_part in range(part._capacity):
        x = line.record_last_track.x[i_part, :]
        hstgm, bin_edges = np.histogram(x,  bins=50, range=(-3, 3), density=True)
        bin_centers = (bin_edges[:-1]+bin_edges[1:])/2
        gauss = np.exp(-bin_centers**2/2)/np.sqrt(2.0*np.pi)
        xo.assert_allclose(hstgm, gauss, rtol=1e-10, atol=1E-2)


@for_all_test_contexts
def test_direct_sampling(test_context):
    n_seeds = 3
    n_samples = 3e6
    ran = xt.RandomNormal(_context=test_context)
    samples = ran.generate(n_samples=n_samples, n_seeds=n_seeds)
    samples = test_context.nparray_from_context_array(samples)

    for i_part in range(n_seeds):
        hstgm, bin_edges = np.histogram(samples[i_part],  bins=50, range=(-3, 3), density=True)
        bin_centers = (bin_edges[:-1]+bin_edges[1:])/2
        gauss = np.exp(-bin_centers**2/2)/np.sqrt(2.0*np.pi)
        xo.assert_allclose(hstgm, gauss, rtol=1e-10, atol=1E-2)


@for_all_test_contexts
def test_reproducibility(test_context):
    # 1e8 samples in total
    n_seeds = int(1e5)
    n_samples_per_seed = int(1e3)
    x_init = np.random.uniform(0.001, 0.003, n_seeds)
    part_init = xp.Particles(x=x_init, p0c=4e11, _context=test_context)
    part_init._init_random_number_generator(seeds=np.arange(n_seeds, dtype=int))
    ran = xt.RandomNormal(_context=test_context)
    part1 = part_init.copy(_context=test_context)
    part2 = part_init.copy(_context=test_context)
    # Instead of having more particles - which would lead to memory issues -
    # we repeatedly sample and compare
    for i in range(12):
        results1 = ran.generate(n_samples=n_samples_per_seed*n_seeds, particles=part1)
        results1 = test_context.nparray_from_context_array(results1)
        results1 = copy.deepcopy(results1)
        results2 = ran.generate(n_samples=n_samples_per_seed*n_seeds, particles=part2)
        results2 = test_context.nparray_from_context_array(results2)
        assert np.all(results1 == results2)




import numpy as np
import xtrack as xt

k2 = 3.
k2s = 5.
length = 0.4

line_thin = xt.Line(elements=[
    xt.Drift(length=length/2),
    xt.Multipole(knl=[0., 0., k2 * length],
                 ksl=[0., 0., k2s * length],
                 length=length),
    xt.Drift(length=length/2),
])
line_thin.build_tracker()

line_thick = xt.Line(elements=[
    xt.Sextupole(k2=k2, k2s=k2s, length=length),
])
line_thick.build_tracker()

p = xt.Particles(
    p0c=6500e9,
    x=[-3e-2, -2e-3, 0, 1e-3, 2e-3, 3e-2],
    px=[1e-6, 2e-6,  0, 2e-6, 1e-6, 1e-6],
    y=[-2e-2, -5e-3, 0, 5e-3, -4e-3, 2e-2],
    py=[2e-6, 4e-6,  0, 2e-6, 1e-6, 1e-6],
    delta=[1e-3, 2e-3, 0, -2e-3, -1e-3, -1e-3],
    zeta=[-5e-2, -6e-3, 0, 6e-3, 5e-3, 5e-2],
)

p_thin = p.copy()
p_thick = p.copy()

line_thin.track(p_thin)
line_thick.track(p_thick)

assert np.allclose(p_thin.x, p_thick.x, rtol=0, atol=1e-14)
assert np.allclose(p_thin.px, p_thick.px, rtol=0, atol=1e-14)
assert np.allclose(p_thin.y, p_thick.y, rtol=0, atol=1e-14)
assert np.allclose(p_thin.py, p_thick.py, rtol=0, atol=1e-14)
assert np.allclose(p_thin.delta, p_thick.delta, rtol=0, atol=1e-14)
assert np.allclose(p_thin.zeta, p_thick.zeta, rtol=0, atol=1e-14)

# slicing
Teapot = xt.slicing.Teapot
Strategy = xt.slicing.Strategy

line_sliced = line_thick.copy()
line_sliced.slice_thick_elements(
    slicing_strategies=[Strategy(slicing=Teapot(5))])
line_sliced.build_tracker()

p_sliced = p.copy()
line_sliced.track(p_sliced)

assert np.allclose(p_sliced.x, p_thick.x, rtol=0, atol=5e-6)
assert np.allclose(p_sliced.px, p_thick.px, rtol=0.01, atol=1e-10)
assert np.allclose(p_sliced.y, p_thick.y, rtol=0, atol=5e-6)
assert np.allclose(p_sliced.py, p_thick.py, rtol=0.01, atol=1e-10)
assert np.allclose(p_sliced.delta, p_thick.delta, rtol=0, atol=1e-14)
assert np.allclose(p_sliced.zeta, p_thick.zeta, rtol=0, atol=2e-7)

line_thin.track(p_thin, backtrack=True)
line_thick.track(p_thick, backtrack=True)
line_sliced.track(p_sliced, backtrack=True)

assert np.allclose(p_thin.x, p.x, rtol=0, atol=1e-14)
assert np.allclose(p_thin.px, p.px, rtol=0, atol=1e-14)
assert np.allclose(p_thin.y, p.y, rtol=0, atol=1e-14)
assert np.allclose(p_thin.py, p.py, rtol=0, atol=1e-14)
assert np.allclose(p_thin.delta, p.delta, rtol=0, atol=1e-14)

assert np.allclose(p_thick.x, p.x, rtol=0, atol=1e-14)
assert np.allclose(p_thick.px, p.px, rtol=0, atol=1e-14)
assert np.allclose(p_thick.y, p.y, rtol=0, atol=1e-14)
assert np.allclose(p_thick.py, p.py, rtol=0, atol=1e-14)
assert np.allclose(p_thick.delta, p.delta, rtol=0, atol=1e-14)

assert np.allclose(p_sliced.x, p.x, rtol=0, atol=1e-14)
assert np.allclose(p_sliced.px, p.px, rtol=0, atol=1e-14)
assert np.allclose(p_sliced.y, p.y, rtol=0, atol=1e-14)
assert np.allclose(p_sliced.py, p.py, rtol=0, atol=1e-14)
assert np.allclose(p_sliced.delta, p.delta, rtol=0, atol=1e-14)
assert np.allclose(p_sliced.zeta, p.zeta, rtol=0, atol=1e-14)

from cpymad.madx import Madx
mad = Madx()
mad.input(f"""
    knob_a := 1.0;
    knob_b := 2.0;
    knob_l := 0.4;
    ss: sequence, l:=2 * knob_b, refer=entry;
        elem: sextupole, at=0, l:=knob_l, k2:=3*knob_a, k2s:=5*knob_b;
    endsequence;
    """)
mad.beam()
mad.use(sequence='ss')

line_mad = xt.Line.from_madx_sequence(mad.sequence.ss, allow_thick=True,
                                      deferred_expressions=True)
line_mad.build_tracker()

elem = line_mad['elem']
assert isinstance(elem, xt.Sextupole)
assert np.isclose(elem.length, 0.4, rtol=0, atol=1e-14)
assert np.isclose(elem.k2, 3, rtol=0, atol=1e-14)
assert np.isclose(elem.k2s, 10, rtol=0, atol=1e-14)

line_mad.vv['knob_a'] = 0.5
line_mad.vv['knob_b'] = 0.6
line_mad.vv['knob_l'] = 0.7

assert np.isclose(elem.length, 0.7, rtol=0, atol=1e-14)
assert np.isclose(elem.k2, 1.5, rtol=0, atol=1e-14)
assert np.isclose(elem.k2s, 3.0, rtol=0, atol=1e-14)




# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

import ducktrack as dtk

context = xo.ContextCpu()

x_aper_min = -0.1
x_aper_max = 0.2
y_aper_min = 0.2
y_aper_max = 0.3

part_gen_range = 0.35
n_part=10000

test_part = dtk.TestParticles(
        p0c=6500e9,
        x=np.random.uniform(-part_gen_range, part_gen_range, n_part),
        px = np.zeros(n_part),
        y=np.random.uniform(-part_gen_range, part_gen_range, n_part),
        py = np.zeros(n_part),
        )

particles = xt.Particles.from_dict(test_part.to_dict(), _context=context)

aper_test = dtk.elements.LimitRect(min_x=x_aper_min,
                                          max_x=x_aper_max,
                                          min_y=y_aper_min,
                                          max_y=y_aper_max)

aper = xt.LimitRect(_context=context,
                    **aper_test.to_dict())
aper.track(particles)

aper_test.track(test_part)

part_id = context.nparray_from_context_array(particles.particle_id)
part_state = context.nparray_from_context_array(particles.state)
part_x = context.nparray_from_context_array(particles.x)
part_y = context.nparray_from_context_array(particles.y)

id_alive = part_id[part_state>0]

assert np.allclose(test_part.particle_id, np.sort(id_alive))

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(part_x, part_y, '.', color='red')
plt.plot(part_x[part_state>0], part_y[part_state>0], '.', color='green')

plt.show()



from cpymad.madx import Madx
import xtrack as xt

mad = Madx()
mad.call('lep.seq9')
mad.call('lep.opt9')
mad.input('beam, particle=positron, energy=100, radiate=true;')
mad.use(sequence='lep')
mad.input('vrfc:=2.; vrfsc:=50.; vrfscn:=50.; ! LEP2  rf on')

twm = mad.twiss()

line = xt.Line.from_madx_sequence(mad.sequence.lep, deferred_expressions=True)
line.particle_ref = xt.Particles(
    mass0=xt.ELECTRON_MASS_EV, gamma0=mad.sequence.lep.beam.gamma)
line.build_tracker()
sv = line.survey()
tw = line.twiss()

# import xplt
# xplt.FloorPlot(sv, line, element_width=100)

# Slice
line.discard_tracker()
line.slice_thick_elements(
    slicing_strategies=[
        # Slicing with thin elements
        xt.Strategy(slicing=xt.Teapot(1)),
        xt.Strategy(slicing=xt.Uniform(2), element_type=xt.Bend),
        xt.Strategy(slicing=xt.Teapot(10), element_type=xt.Quadrupole),
    ])
line.build_tracker()
line.configure_radiation('mean')

tw_rad = line.twiss(eneloss_and_damping=True)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
sp1 = plt.subplot(3, 1, 1)
plt.plot(tw_rad.s, tw_rad.betx, label='betx')
plt.plot(tw_rad.s, tw_rad.bety, label='bety')
plt.ylabel(r'$\beta_{x,y}$ [m]')

sp2 = plt.subplot(3, 1, 2, sharex=sp1)
plt.plot(tw_rad.s, tw_rad.x, label='x')
plt.ylabel('x [m]')

sp3 = plt.subplot(3, 1, 3, sharex=sp1)
plt.plot(tw_rad.s, tw_rad.delta, label='delta')
plt.ylabel(r'$\delta$')
plt.xlabel('s [m]')

plt.show()


import xtrack as xt
import numpy as np
import pytest

from xobjects.test_helpers import for_all_test_contexts
import xobjects as xo
from cpymad.madx import Madx

assert_allclose = np.testing.assert_allclose

@for_all_test_contexts
@pytest.mark.parametrize(
    'slice_mode',
    [None, 'thin', 'thick'],
    ids=['no_slice', 'thin_slice', 'thick_slice'])
def test_test_tilt_shifts_vs_sandwtch(test_context, slice_mode):

    ele_test = [
        xt.Bend(k0=0.04, h=0.03, length=1,
                k1=0.1,
                knl=[0.7, 0.8, 0.9, 1.0], ksl=[0.1, 0.2, 0.3, 0.4],
                edge_entry_angle=0.05, edge_exit_angle=0.06,
                edge_entry_hgap=0.06, edge_exit_hgap=0.07,
                edge_entry_fint=0.08, edge_exit_fint=0.09,
                ),
        xt.Quadrupole(k1=2., k1s=-3., length=3.),
        xt.Sextupole(k2=0.1, k2s=0.2, length=0.3),
        xt.Octupole(k3=0.1, k3s=0.2, length=0.4),
        xt.Multipole(knl=[0.7, 0.8, 0.9, 1.0], ksl=[0.1, 0.2, 0.3, 0.4],
                            length=0.4, hxl=0.1)
    ]

    for elem in ele_test:
        print('ele type:', elem.__class__.__name__)

        shift_x = 1e-3
        shift_y = 2e-3
        shift_s = 10e-3
        rot_s_rad = -0.4

        line_test = xt.Line(elements=[elem.copy()])

        line_ref = xt.Line(elements=[
            xt.Drift(length=shift_s),
            xt.XYShift(dx=shift_x, dy=shift_y),
            xt.SRotation(angle=np.rad2deg(rot_s_rad)),
            elem.copy(),
            xt.SRotation(angle=np.rad2deg(-rot_s_rad)),
            xt.XYShift(dx=-shift_x, dy=-shift_y),
            xt.Drift(length=-shift_s)
        ])
        line_ref.config.XTRACK_GLOBAL_XY_LIMIT = 1000

        if slice_mode is not None:
            line_test.slice_thick_elements(
                slicing_strategies=[xt.Strategy(xt.Teapot(3, mode=slice_mode))])
            line_ref.slice_thick_elements(
                slicing_strategies=[xt.Strategy(xt.Teapot(3, mode=slice_mode))])

        line_test['e0'].rot_s_rad = rot_s_rad
        line_test['e0'].shift_x = shift_x
        line_test['e0'].shift_y = shift_y
        line_test['e0'].shift_s = shift_s

        p_test = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03,
                              _context=test_context)
        p_ref = p_test.copy()
        p0 = p_test.copy()

        line_test.build_tracker(_context=test_context)
        line_ref.build_tracker(_context=test_context)

        line_test.track(p_test)
        line_ref.track(p_ref)

        p_test.move(_context=xo.context_default)
        p_ref.move(_context=xo.context_default)

        assert_allclose = np.testing.assert_allclose
        assert_allclose(p_test.x, p_ref.x, rtol=0, atol=1e-13)
        assert_allclose(p_test.px, p_ref.px, rtol=0, atol=1e-13)
        assert_allclose(p_test.y, p_ref.y, rtol=0, atol=1e-13)
        assert_allclose(p_test.py, p_ref.py, rtol=0, atol=1e-13)
        assert_allclose(p_test.zeta, p_ref.zeta, rtol=0, atol=5e-12)
        assert_allclose(p_test.delta, p_ref.delta, rtol=0, atol=1e-13)

        # Test backtrack
        p_test.move(_context=test_context)
        p_ref.move(_context=test_context)

        line_test.track(p_test, backtrack=True)
        line_ref.track(p_ref, backtrack=True)

        p_test.move(_context=xo.context_default)
        p_ref.move(_context=xo.context_default)

        assert_allclose(p_test.x, p_ref.x, rtol=0, atol=1e-11)
        assert_allclose(p_test.px, p_ref.px, rtol=0, atol=1e-11)
        assert_allclose(p_test.y, p_ref.y, rtol=0, atol=1e-11)
        assert_allclose(p_test.py, p_ref.py, rtol=0, atol=1e-11)
        assert_allclose(p_test.zeta, p_ref.zeta, rtol=0, atol=1e-7)
        assert_allclose(p_test.delta, p_ref.delta, rtol=0, atol=1e-11)

        p0.move(_context=xo.context_default)
        assert_allclose(p_test.x, p0.x, rtol=0, atol=1e-11)
        assert_allclose(p_test.px, p0.px, rtol=0, atol=1e-11)
        assert_allclose(p_test.y, p0.y, rtol=0, atol=1e-11)
        assert_allclose(p_test.py, p0.py, rtol=0, atol=1e-11)
        assert_allclose(p_test.zeta, p0.zeta, rtol=0, atol=1e-7)
        assert_allclose(p_test.delta, p0.delta, rtol=0, atol=1e-11)

def test_tilt_shifts_vs_madx():

    mad = Madx()

    tilt_deg = 12
    shift_x = -2e-3
    shift_y = 3e-3
    k1 = 0.2
    tilt_rad = np.deg2rad(tilt_deg)

    x_test = 1e-3
    px_test = 2e-3
    y_test = 3e-3
    py_test = 4e-3

    mad.input(f"""
    k1={k1};

    elm: quadrupole,
        k1:=k1,
        l=1,
        tilt={tilt_rad};

    seq: sequence, l=1;
    elm: elm, at=0.5;
    endsequence;

    beam, particle=proton, gamma=100;
    use, sequence=seq;

    select,flag=error,clear;
    select,flag=error,pattern=elm;
    ealign, dx={shift_x}, dy={shift_y};

    twiss, betx=1, bety=1, x={x_test}, px={px_test}, y={y_test}, py={py_test};

    """)

    elm = xt.Quadrupole(k1=k1, length=1)

    elm_tilted = xt.Quadrupole(k1=k1, length=1, rot_s_rad=tilt_rad,
                            shift_x=shift_x, shift_y=shift_y)

    lsandwitch = xt.Line(elements=[
        xt.XYShift(dx=shift_x, dy=shift_y),
        xt.SRotation(angle=tilt_deg),
        elm,
        xt.SRotation(angle=-tilt_deg),
        xt.XYShift(dx=-shift_x, dy=-shift_y)
    ])
    lsandwitch.build_tracker()

    l_tilted = xt.Line(elements=[elm_tilted])
    l_tilted.build_tracker()

    lmad = xt.Line.from_madx_sequence(mad.sequence.seq, enable_align_errors=True)
    lmad.build_tracker()

    p0 = xt.Particles(x=x_test, px=px_test, y=y_test, py=py_test, gamma0=100)

    pmad = p0.copy()
    lmad.track(pmad)

    psandwitch = p0.copy()
    lsandwitch.track(psandwitch)

    plinetilted = p0.copy()
    l_tilted.track(plinetilted)

    peletitled = p0.copy()
    elm_tilted.track(peletitled)

    assert elm.rot_s_rad == 0
    elm.rot_s_rad = tilt_rad
    elm.shift_x = shift_x
    elm.shift_y = shift_y
    pprop = p0.copy()
    elm.track(pprop)

    for pp in [psandwitch, plinetilted, pmad, peletitled, pprop]:
        assert_allclose(pp.x, mad.table.twiss.x[-1], rtol=0, atol=1e-12)
        assert_allclose(pp.px, mad.table.twiss.px[-1], rtol=0, atol=1e-12)
        assert_allclose(pp.y, mad.table.twiss.y[-1], rtol=0, atol=1e-12)
        assert_allclose(pp.py, mad.table.twiss.py[-1], rtol=0, atol=1e-12)
        assert_allclose(pp.zeta, pp.beta0[0]*mad.table.twiss.t[-1], rtol=0, atol=1e-12)


@for_all_test_contexts
def test_shift_x(test_context):

    k1 = 2.
    length = 0.1

    quad = xt.Quadrupole(k1=k1, length=length, shift_x=1e-3, _context=test_context)

    assert quad.shift_x == 1e-3
    assert quad.shift_y == 0
    assert quad._sin_rot_s == 0.0
    assert quad._cos_rot_s == 1.0

    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    quad.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.px, -k1 * length * -1e-3, rtol=5e-3, atol=0)

    # Change the shift
    quad.shift_x = 2e-3
    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    quad.track(p)
    p.move(_context=xo.context_default)
    assert_allclose(p.px, -k1 * length * -2e-3, rtol=5e-3, atol=0)

    # Make a line
    line = xt.Line(elements=[quad])

    # Slice the line:
    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(3))])
    line.build_tracker(_context=test_context)

    tt = line.get_table()
    assert len(tt.rows[r'e0\.\..*']) == 5

    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    line.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.px, -k1 * length * -2e-3, rtol=5e-3, atol=0)

    # Change the shift
    quad.shift_x = 3e-3
    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    line.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.px, -k1 * length * -3e-3, rtol=5e-3, atol=0)

@for_all_test_contexts
def test_shift_y(test_context):

    k1 = 2.
    length = 0.1

    quad = xt.Quadrupole(k1=k1, length=length, shift_y=1e-3, _context=test_context)

    assert quad.shift_x == 0
    assert quad.shift_y == 1e-3
    assert quad._sin_rot_s == 0.0
    assert quad._cos_rot_s == 1.0

    p = xt.Particles(y=0, p0c=1e12, _context=test_context)
    quad.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.py, k1 * length * -1e-3, rtol=5e-3, atol=0)

    # Change the shift
    quad.shift_y = 2e-3
    p = xt.Particles(y=0, p0c=1e12, _context=test_context)
    quad.track(p)
    p.move(_context=xo.context_default)
    assert_allclose(p.py, k1 * length * -2e-3, rtol=5e-3, atol=0)

    # Make a line
    line = xt.Line(elements=[quad])

    # Slice the line:
    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(3))])
    line.build_tracker(_context=test_context)

    tt = line.get_table()
    assert len(tt.rows[r'e0\.\..*']) == 5

    p = xt.Particles(y=0, p0c=1e12, _context=test_context)
    line.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.py, k1 * length * -2e-3, rtol=5e-3, atol=0)

    # Change the shift
    quad.shift_y = 3e-3
    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    line.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.py, k1 * length * -3e-3, rtol=5e-3, atol=0)

@for_all_test_contexts
def test_rot_s(test_context):

    k0 = 2.
    length = 0.1
    rot_s_rad = 0.2

    bend = xt.Bend(k0=k0, length=length, rot_s_rad=rot_s_rad, _context=test_context)

    assert bend.shift_x == 0
    assert bend.shift_y == 0
    assert_allclose(bend._sin_rot_s, np.sin(rot_s_rad), rtol=0, atol=1e-14)
    assert_allclose(bend._cos_rot_s, np.cos(rot_s_rad), rtol=0, atol=1e-14)

    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    bend.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3, atol=0)
    assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3, atol=0)

    rot_s_rad = 0.3
    bend.rot_s_rad = rot_s_rad
    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    bend.track(p)
    p.move(_context=xo.context_default)
    assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3, atol=0)
    assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3, atol=0)

    # Make a line
    line = xt.Line(elements=[bend])

    # Slice the line:
    line.slice_thick_elements(
        slicing_strategies=[xt.Strategy(xt.Uniform(3))])
    line.build_tracker(_context=test_context)

    tt = line.get_table()
    assert len(tt.rows[r'e0\.\..?']) == 3

    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    line.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3,
                    atol=0)
    assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3,
                    atol=0)

    rot_s_rad = 0.4
    bend.rot_s_rad = rot_s_rad
    p = xt.Particles(x=0, p0c=1e12, _context=test_context)
    line.track(p)
    p.move(_context=xo.context_default)

    assert_allclose(p.px, -k0 * length * np.cos(rot_s_rad), rtol=5e-3,
                    atol=0)
    assert_allclose(p.py, -k0 * length * np.sin(rot_s_rad), rtol=5e-3,
                    atol=0)



import itertools

from cpymad.madx import Madx

mad = Madx()
mad.call("multipole.madx")

from xtrack import MadLoader

def gen_options(opt1):
    for v in itertools.product(*([[False,True]]*len(opt1))):
        yield dict(zip(opt1,v))

opt1=["enable_expressions",
      "enable_errors",
      "enable_apertures"]

opt2=["skip_markers","merge_drifts","merge_multipoles"]


for opt in gen_options(opt1):
    print(opt)
    ml=MadLoader(mad.sequence.seq,**opt)
    line=ml.make_line()
    print(line.element_names)
    print()

for opt in gen_options(opt2):
    print(opt)
    ml=MadLoader(mad.sequence.seq,**opt)
    line=list(ml.iter_elements())
    print(line)
    print()






import xtrack as xt
import numpy as np

mm = xt.Multipole(knl=[0.001, 0.001, 0.01, 0.02, 0.04, 0.6],
                  hxl=0.002, length=2.)

lst = [mm.copy() for i in range(10000)]

p = xt.Particles(p0c=7e12, x=np.linspace(-1e-3, 1e-3, 1000))

line = xt.Line(elements=lst)
line.track(p)

line.build_tracker()
line.track(p, num_turns=100, with_progress=True, time=True)
print('Time for tracking: ', line.time_last_track)


import xtrack as xt

# Load a line and build a tracker
line = xt.Line.from_json('../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
line.build_tracker()

opt = line.match(
    start='mq.30l8.b1', end='mq.23l8.b1',
    betx=1, bety=1, y=0, py=0, # <-- conditions at start
    vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1'],
                    step=1e-10, limits=[-1e-3, 1e-3]),
    targets = [
        xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1'),
        xt.TargetSet(y=0, py=0, at=xt.END)
    ])

opt.target_status()
# prints:
#
# Target status:
# id state tag tol_met      residue  current_val target_val description
#  0 ON           True  1.30104e-18        0.003      0.003 ('y', 'mb.b28l8.b1'), val=0.003, tol=1e- ...
#  1 ON           True -3.38813e-20 -3.38813e-20          0 ('py', 'mb.b28l8.b1'), val=0, tol=1e-10, ...
#  2 ON           True   -4.127e-17   -4.127e-17          0 ('y', 'mq.23l8.b1'), val=0, tol=1e-10, w ...
#  3 ON           True  -6.1664e-19  -6.1664e-19          0 ('py', 'mq.23l8.b1'), val=0, tol=1e-10,  ...

#!end-doc-part

import matplotlib.pyplot as plt

opt.tag(tag='matched')
opt.reload(0)
tw_before = line.twiss(method='4d')
opt.reload(tag='matched')
tw = line.twiss(method='4d')

plt.close('all')
fig = plt.figure(1, figsize=(6.4*1.2, 4.8*0.8))
ax = fig.add_subplot(111)
ax.plot(tw.s, tw.y*1000, label='y')

for nn in ['mcbv.30l8.b1', 'mcbv.28l8.b1', 'mcbv.26l8.b1', 'mcbv.24l8.b1']:
    ax.axvline(x=line.get_s_position(nn), color='k', linestyle='--', alpha=0.5)
    ax.text(line.get_s_position(nn), 10, nn, rotation=90,
            horizontalalignment='left', verticalalignment='top')

ax.axvline(x=line.get_s_position('mb.b28l8.b1'), color='r', linestyle='--', alpha=0.5)
ax.text(line.get_s_position('mb.b28l8.b1'), 10, 'mb.b28l8.b1', rotation=90,
        horizontalalignment='left', verticalalignment='top')

ax.axvline(x=line.get_s_position('mq.30l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.axvline(x=line.get_s_position('mq.23l8.b1'), color='g', linestyle='--', alpha=0.5)
ax.text(line.get_s_position('mq.30l8.b1'), 10, 'mq.30l8.b1', rotation=90,
        horizontalalignment='right', verticalalignment='top')
ax.text(line.get_s_position('mq.23l8.b1'), 10, 'mq.23l8.b1', rotation=90,
        horizontalalignment='right', verticalalignment='top')

ax.set_xlim(line.get_s_position('mq.30l8.b1') - 10,
            line.get_s_position('mq.23l8.b1') + 10)
ax.set_xlabel('s [m]')
ax.set_ylabel('y [mm]')
ax.set_ylim(-0.5, 10)
plt.subplots_adjust(bottom=.152, top=.9, left=.1, right=.95)
plt.show()

import numpy as np
assert np.isclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
assert np.isclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
assert np.isclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
assert np.isclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
assert np.isclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)



import numpy as np
import pandas as pd

import xtrack as xt

from xtrack import PipelineStatus

# For example we build an element type that puts the simulation on hold based on
# an internal counter.

class DummyPipelinedElement:
    def __init__(self, n_hold=0):
        self.n_hold = n_hold
        self.iscollective = True
        self.i_hold = 0

    def track(self, particles):
        self.i_hold += 1
        if self.i_hold < self.n_hold:
            return PipelineStatus(on_hold=True,
                info=f'stopped by internal counter {self.i_hold}/{self.n_hold}')
        else:
            self.i_hold = 0

# We build two trackers, each with an element that puts the simulation on hold
line1 = xt.Line(elements=[xt.Drift(length=1),
                           DummyPipelinedElement(n_hold=2),
                           xt.Drift(length=1)],
                 element_names=['d11', 'pipelnd_el1', 'd12'])
line1.build_tracker(enable_pipeline_hold=True)
line1.reset_s_at_end_turn = False

line2 = xt.Line(elements=[xt.Drift(length=1),
                           DummyPipelinedElement(n_hold=3),
                           xt.Drift(length=1)],
                element_names=['d11', 'pipelnd_el2', 'd12'])
line2.build_tracker(enable_pipeline_hold=True)
line2.reset_s_at_end_turn = False

# We use a multitracker to track one two particles with the first tracker
# and one particle with the second one

p1 = xt.Particles(p0c=7e12, x=[0,0,0])
p2 = xt.Particles(p0c=7e12, x=[0,0,0])
p3 = xt.Particles(p0c=7e12, x=[0,0,0])

multitracker = xt.PipelineMultiTracker(
    branches=[xt.PipelineBranch(line=line1, particles=p1),
              xt.PipelineBranch(line=line1, particles=p2),
              xt.PipelineBranch(line=line2, particles=p3),
              ],
    enable_debug_log=True)

multitracker.track(num_turns=4)

# The simulation switched multiple times from one brach to the other (due to the
# pipelined elements as it can be seen by the debug log (it is convenient to use
# a pandas dataframe to view the debug log).

log_df = pd.DataFrame(multitracker.debug_log)

# Contains:
#     branch  track_session_turn held_by_element                             info
# 0        0                   0     pipelnd_el1  stopped by internal counter 1/2
# 1        1                   1     pipelnd_el1  stopped by internal counter 1/2
# 2        2                   0     pipelnd_el2  stopped by internal counter 1/3
# 3        0                   1     pipelnd_el1  stopped by internal counter 1/2
# 4        1                   2     pipelnd_el1  stopped by internal counter 1/2
# 5        2                   0     pipelnd_el2  stopped by internal counter 2/3
# 6        0                   2     pipelnd_el1  stopped by internal counter 1/2
# [...]

# Check that all particles have been tracked for 4 turns
assert np.all(p1.at_turn == 4)
assert np.all(p2.at_turn == 4)
assert np.all(p3.at_turn == 4)
assert np.all(p1.s == 8.)
assert np.all(p2.s == 8.)
assert np.all(p3.s == 8.)

############  A simpler case for some extra checks

# Reset the counters
line1['pipelnd_el1'].i_hold = 0
line2['pipelnd_el2'].i_hold = 0

p1 = xt.Particles(p0c=7e12, x=[0,0,0])
p2 = xt.Particles(p0c=7e12, x=[0,0,0])

multitracker = xt.PipelineMultiTracker(
    branches=[xt.PipelineBranch(line=line1, particles=p1),
              xt.PipelineBranch(line=line2, particles=p2),
              ],
    enable_debug_log=True)

multitracker.track(num_turns=4)

# The debug log can be loaded in a dataframe
log_df = pd.DataFrame(multitracker.debug_log)

# Looks like this
#     branch  track_session_turn held_by_element                             info
# 0        0                   0     pipelnd_el1  stopped by internal counter 1/2
# 1        1                   0     pipelnd_el2  stopped by internal counter 1/3
# 2        0                   1     pipelnd_el1  stopped by internal counter 1/2
# 3        1                   0     pipelnd_el2  stopped by internal counter 2/3
# 4        0                   2     pipelnd_el1  stopped by internal counter 1/2
# 5        1                   1     pipelnd_el2  stopped by internal counter 1/3
# 6        0                   3     pipelnd_el1  stopped by internal counter 1/2
# 7        1                   1     pipelnd_el2  stopped by internal counter 2/3
# 8        1                   2     pipelnd_el2  stopped by internal counter 1/3
# 9        1                   2     pipelnd_el2  stopped by internal counter 2/3
# 10       1                   3     pipelnd_el2  stopped by internal counter 1/3
# 11       1                   3     pipelnd_el2  stopped by internal counter 2/3

# Some checks

assert len(multitracker.debug_log) == 12
assert p1.at_turn[0] == 4
assert p2.at_turn[0] == 4
assert p1.s[0] == 8.
assert p2.s[0] == 8.
assert np.all(log_df['branch'] == np.array(4*[0, 1] + 4*[1]))
assert np.all(log_df[log_df['branch']==0]['held_by_element'] == 'pipelnd_el1')
assert np.all(log_df[log_df['branch']==1]['held_by_element'] == 'pipelnd_el2')





from cpymad.madx import Madx
import xtrack as xt

test_data_folder = '../../test_data/'
mad = Madx()

mad.call(test_data_folder + 'pimms/PIMM_orig.seq')
mad.call(test_data_folder + 'pimms/betatron.str')
mad.beam(particle='proton', gamma=1.21315778) # 200 MeV
mad.use('pimms')
seq = mad.sequence.pimms
def_expr = True

line = xt.Line.from_madx_sequence(seq, deferred_expressions=def_expr)
line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
                                 mass0=seq.beam.mass * 1e9,
                                 q0=seq.beam.charge)

sv = line.survey()

tt = line.get_table()
ttsext = tt.rows[tt.element_type == 'Sextupole']

import xplt
import matplotlib.pyplot as plt
plt.close('all')
xplt.FloorPlot(sv, line, labels=ttsext.name)

plt.show()


import numpy as np
import matplotlib.pyplot as plt

import xtrack as xt
import xobjects as xo
import xpart as xp

import RF_Track as RFT

p0c = 1.2e9  # eV
Ld = 0.2 # m

#############################################
#######  RF-Track's part starts here  #######
#############################################

# Define the RF-Track element
dft = RFT.Drift(Ld)
dft.set_aperture(0.08, 0.02, 'rectangular')

#############################################
#######  RF-Track's part ends here    #######
#############################################

# Back to Xsuite
pi = np.pi
lbend = 3
elements = {
    'a1': xt.LimitRect(min_x=-0.02, max_x=0.02, min_y=-0.08, max_y=0.08),
    'd1': xt.Drift(length=Ld),
    'd2': xt.RFT_Element(element=dft),
}

# Build the ring
line = xt.Line(elements=elements, element_names=list(elements.keys()))
line.particle_ref = xt.Particles(p0c=p0c, mass0=xt.PROTON_MASS_EV)
line.configure_bend_model(core='full', edge=None)

# track
## Choose a context
context = xo.ContextCpu()         # For CPU
# context = xo.ContextCupy()      # For CUDA GPUs
# context = xo.ContextPyopencl()  # For OpenCL GPUs

## Transfer lattice on context and compile tracking code
line.build_tracker(_context=context)

## Build particle object on context
n_part = 10000

rng = np.random.default_rng(12345)

particles = xp.Particles(p0c=p0c, #eV
                         q0=1, mass0=xp.PROTON_MASS_EV,
                         x=rng.uniform(-0.10, 0.10, n_part),
                         y=rng.uniform(-0.10, 0.10, n_part),
                         px=np.zeros(n_part),
                         py=np.zeros(n_part),
                         zeta=np.zeros(n_part),
                         delta=np.zeros(n_part),
                         _context=context)

print('tracking starts')
line.track(particles)
print('tracking ends')

plt.figure(1)
plt.scatter(particles.x*1e3, particles.y*1e3, s=10, facecolors='none', edgecolors='b')
plt.xlabel("$x$ [mm]")
plt.ylabel("$y$ [mm]")
plt.xlim([-80, 80])
plt.ylim([-80, 80])
plt.show()




import numpy as np
import xtrack as xt
import xobjects as xo
from scipy.constants import c as clight
from scipy.constants import e as qe

from cpymad.madx import Madx

fname = 'fccee_z'; pc_gev = 45.6
# fname = 'fccee_t'; pc_gev = 182.5

mad = Madx()
mad.call('../../test_data/fcc_ee/' + fname + '.seq')
mad.beam(particle='positron', pc=pc_gev)
mad.use('fccee_p_ring')

line = xt.Line.from_madx_sequence(mad.sequence.fccee_p_ring, allow_thick=True,
                                  deferred_expressions=True)
line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                                 gamma0=mad.sequence.fccee_p_ring.beam.gamma)
line.cycle('ip.4', inplace=True)
line.append_element(element=xt.Marker(), name='ip.4.l')

Strategy = xt.Strategy
Teapot = xt.Teapot
slicing_strategies = [
    Strategy(slicing=None),  # Default catch-all as in MAD-X
    Strategy(slicing=Teapot(3), element_type=xt.Bend),
    Strategy(slicing=Teapot(3), element_type=xt.CombinedFunctionMagnet),
    # Strategy(slicing=Teapot(50), element_type=xt.Quadrupole), # Starting point
    Strategy(slicing=Teapot(5), name=r'^qf.*'),
    Strategy(slicing=Teapot(5), name=r'^qd.*'),
    Strategy(slicing=Teapot(5), name=r'^qfg.*'),
    Strategy(slicing=Teapot(5), name=r'^qdg.*'),
    Strategy(slicing=Teapot(5), name=r'^ql.*'),
    Strategy(slicing=Teapot(5), name=r'^qs.*'),
    Strategy(slicing=Teapot(10), name=r'^qb.*'),
    Strategy(slicing=Teapot(10), name=r'^qg.*'),
    Strategy(slicing=Teapot(10), name=r'^qh.*'),
    Strategy(slicing=Teapot(10), name=r'^qi.*'),
    Strategy(slicing=Teapot(10), name=r'^qr.*'),
    Strategy(slicing=Teapot(10), name=r'^qu.*'),
    Strategy(slicing=Teapot(10), name=r'^qy.*'),
    Strategy(slicing=Teapot(50), name=r'^qa.*'),
    Strategy(slicing=Teapot(50), name=r'^qc.*'),
    Strategy(slicing=Teapot(20), name=r'^sy\..*'),
    Strategy(slicing=Teapot(30), name=r'^mwi\..*'),
]
line.discard_tracker()
line.slice_thick_elements(slicing_strategies=slicing_strategies)
line.build_tracker()

theta_tilt = 15e-3 # rad
l_solenoid = 4.4
l_beam = l_solenoid / np.cos(theta_tilt)
ds_sol_start = -l_beam / 2
ds_sol_end = +l_beam / 2
ip_sol = 'ip.1'


tt = line.get_table(attr=True)

s_ip = tt['s', ip_sol]

line.discard_tracker()
line.insert_element(name='sol_start_'+ip_sol, element=xt.Marker(),
                    at_s=s_ip + ds_sol_start)
line.insert_element(name='sol_end_'+ip_sol, element=xt.Marker(),
                    at_s=s_ip + ds_sol_end)

sol_start_tilt = xt.YRotation(angle=-theta_tilt * 180 / np.pi)
sol_end_tilt = xt.YRotation(angle=+theta_tilt * 180 / np.pi)
sol_start_shift = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt))
sol_end_shift = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt))

line.element_dict['sol_start_tilt_'+ip_sol] = sol_start_tilt
line.element_dict['sol_end_tilt_'+ip_sol] = sol_end_tilt
line.element_dict['sol_start_shift_'+ip_sol] = sol_start_shift
line.element_dict['sol_end_shift_'+ip_sol] = sol_end_shift

line.element_dict['sol_entry_'+ip_sol] = xt.Marker()
line.element_dict['sol_exit_'+ip_sol] = xt.Marker()
s_sol_slices = np.linspace(-l_solenoid/2, l_solenoid/2, 11)
l_sol_slices = np.diff(s_sol_slices)
s_sol_slices_entry = s_sol_slices[:-1]

sol_slices = []
for ii in range(len(s_sol_slices_entry)):
    # sol_slices.append(xt.Solenoid(length=l_sol_slices[ii], ks=0)) # Off for now
    sol_slices.append(xt.Drift(length=l_sol_slices[ii])) # Off for now

sol_slice_names = []
sol_slice_names.append('sol_entry_'+ip_sol)
for ii in range(len(s_sol_slices_entry)):
    nn = f'sol_slice_{ii}_{ip_sol}'
    line.element_dict[nn] = sol_slices[ii]
    sol_slice_names.append(nn)
sol_slice_names.append('sol_exit_'+ip_sol)

tt = line.get_table()
names_upstream = list(tt.rows[:'sol_start_'+ip_sol].name)
names_downstream = list(tt.rows['sol_end_'+ip_sol:].name[:-1]) # -1 to exclude '_end_point' added by the table

element_names = (names_upstream
                 + ['sol_start_tilt_'+ip_sol, 'sol_start_shift_'+ip_sol]
                 + sol_slice_names
                 + ['sol_end_shift_'+ip_sol, 'sol_end_tilt_'+ip_sol]
                 + names_downstream)

line.element_names = element_names

line.config.XTRACK_USE_EXACT_DRIFTS = True
line.build_tracker()

line.configure_radiation(model='mean')
line.compensate_radiation_energy_loss()

tw = line.twiss(eneloss_and_damping=True, particle_on_co=line.particle_ref.copy())

print(tw.partition_numbers)